# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import pack_int
from typing import Literal, Optional

class MsgSvRaceFinish(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            client_id: int = 0,
            time: int = 0,
            diff: int = 0,
            record_personal: bool = False,
            record_server: bool = False
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'sv_race_finish'
        self.system_message: bool = False
        self.message_id: int = 35
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.7')
        self.header: ChunkHeader = chunk_header

        self.client_id: int = client_id
        self.time: int = time
        self.diff: int = diff
        self.record_personal: bool = record_personal
        self.record_server: bool = record_server

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'client_id', self.client_id
        yield 'time', self.time
        yield 'diff', self.diff
        yield 'record_personal', self.record_personal
        yield 'record_server', self.record_server

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.client_id = unpacker.get_int()
        self.time = unpacker.get_int()
        self.diff = unpacker.get_int()
        self.record_personal = unpacker.get_int() == 1
        self.record_server = unpacker.get_int() == 1
        return True

    def pack(self) -> bytes:
        return pack_int(self.client_id) + \
            pack_int(self.time) + \
            pack_int(self.diff) + \
            pack_int(self.record_personal) + \
            pack_int(self.record_server)


# connless
# pylint: disable=duplicate-code

CONNLESS_HEARTBEAT = b'\xff\xff\xff\xffbea2'
CONNLESS_REQUEST_LIST = b'\xff\xff\xff\xffreq2'
CONNLESS_LIST = b'\xff\xff\xff\xfflis2'
CONNLESS_REQUEST_COUNT = b'\xff\xff\xff\xffcou2'
CONNLESS_COUNT = b'\xff\xff\xff\xffsiz2'
CONNLESS_REQUEST_INFO = b'\xff\xff\xff\xffgie3'
CONNLESS_INFO = b'\xff\xff\xff\xffinf3'
CONNLESS_FORWARD_CHECK = b'\xff\xff\xff\xfffw??'
CONNLESS_FORWARD_RESPONSE = b'\xff\xff\xff\xfffw!!'
CONNLESS_FORWARD_OK = b'\xff\xff\xff\xfffwok'
CONNLESS_FORWARD_ERROR = b'\xff\xff\xff\xfffwer'

# control

CTRL_KEEPALIVE = 0
CTRL_CONNECT = 1
CTRL_ACCEPT = 2
# yes control message 3 is missing in 0.7
CTRL_CLOSE = 4
CTRL_TOKEN = 5

# system
NULL = 0
INFO = 1
MAP_CHANGE = 2 # sent when client should switch map
MAP_DATA = 3   # map transfer, contains a chunk of the map file
SERVER_INFO = 4
CON_READY = 5  # connection is ready, client should send start info
SNAP = 6       # normal snapshot, multiple parts
SNAP_EMPTY = 7  # empty snapshot
SNAP_SINGLE = 8 # ?
SNAP_SMALL = 9
INPUT_TIMING = 10   # reports how off the input was
RCON_AUTH_ON = 11  # rcon authentication enabled
RCON_AUTH_OFF = 12 # rcon authentication disabled
RCON_LINE = 13     # line that should be printed to the remote console
RCON_CMD_ADD = 14
RCON_CMD_REM = 15
AUTH_CHALLENGE = 16
AUTH_RESULT = 17
READY = 18
ENTER_GAME = 19
INPUT = 20
RCON_CMD = 21
RCON_AUTH = 22
REQUEST_MAP_DATA = 23
AUTH_START = 24
AUTH_RESPONSE = 25
PING = 26
PING_REPLY = 27
ERROR = 28
MAPLIST_ENTRY_ADD = 29
MAPLIST_ENTRY_REM = 30

# game
INVALID = 0
SV_MOTD = 1
SV_BROADCAST = 2
SV_CHAT = 3
SV_TEAM = 4
SV_KILL_MSG = 5
SV_TUNE_PARAMS = 6
SV_EXTRA_PROJECTILE = 7
SV_READY_TO_ENTER = 8
SV_WEAPON_PICKUP = 9
SV_EMOTICON = 10
SV_VOTE_CLEAR_OPTIONS = 11
SV_VOTE_OPTION_LIST_ADD = 12
SV_VOTE_OPTION_ADD = 13
SV_VOTE_OPTION_REMOVE = 14
SV_VOTE_SET = 15
SV_VOTE_STATUS = 16
SV_SERVER_SETTINGS = 17
SV_CLIENT_INFO = 18
SV_GAME_INFO = 19
SV_CLIENT_DROP = 20
SV_GAME_MSG = 21
DE_CLIENT_ENTER = 22
DE_CLIENT_LEAVE = 23
CL_SAY = 24
CL_SET_TEAM = 25
CL_SET_SPECTATOR_MODE = 26
CL_START_INFO = 27
CL_KILL = 28
CL_READY_CHANGE = 29
CL_EMOTICON = 30
CL_VOTE = 31
CL_CALL_VOTE = 32
SV_SKIN_CHANGE = 33
CL_SKIN_CHANGE = 34
SV_RACE_FINISH = 35
SV_CHECKPOINT = 36
SV_COMMAND_INFO = 37
SV_COMMAND_INFO_REMOVE = 38
CL_COMMAND = 39
NUM_GAME_MESSAGES = 40



class PrettyPrint():
    def __repr__(self):
        return "<class: '" + str(self.__class__.__name__) + "'>"
    def __str__(self):
        return "<class: '" + str(self.__class__.__name__) + "'>: " + str(self.__dict__)




# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjPlayerInfo(PrettyPrint):
    def __init__(
            self,
            player_flags: int = 0,
            score: int = 0,
            latency: int = 0
    ) -> None:
        self.item_name: str = 'item.player_info'
        self.type_id: int = 11
        self.id: int = 0
        self.size: int = 3

        self.player_flags: int = player_flags
        self.score: int = score
        self.latency: int = latency

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'player_flags', self.player_flags
        yield 'score', self.score
        yield 'latency', self.latency

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'player_flags': self.player_flags,
            'score': self.score,
            'latency': self.latency
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.player_flags = unpacker.get_int() # TODO: this is a flag
        self.score = unpacker.get_int()
        self.latency = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.player_flags) + \
            pack_int(self.score) + \
            pack_int(self.latency)


class ParserError(Exception):
    """
    This error is thrown when there is an error
    unpacking teeworlds traffic
    """

class PackerError(Exception):
    """
    This error is thrown when there is an error
    packing teeworlds traffic
    """



from typing import Optional

import twnet_parser.msg6
from twnet_parser.ctrl_message import CtrlMessage

import twnet_parser.messages6.control.keep_alive as keep_alive6
import twnet_parser.messages6.control.connect as connect6
import twnet_parser.messages6.control.accept_connect as accept_connect6
import twnet_parser.messages6.control.ack_accept_connect as ack_accept_connect6
import twnet_parser.messages6.control.close as close6
from twnet_parser.packer import Unpacker
from twnet_parser.packet6 import TwPacket6

def match_control6(msg_id: int, unpacker: Unpacker, packet: TwPacket6, client: bool) -> CtrlMessage:
    msg: Optional[CtrlMessage] = None

    if msg_id == twnet_parser.msg6.CTRL_KEEPALIVE:
        msg = keep_alive6.CtrlKeepAlive()
    elif msg_id == twnet_parser.msg6.CTRL_CONNECT:
        # TODO: this does not belong here
        if unpacker.remaining_size() == 0:
            packet.set_version('0.6.4')
            msg = connect6.CtrlConnect64()
        elif unpacker.data()[0:4] == b'TKEN':
            packet.is_ddnet = True
            packet.set_version('0.6.4')
            msg = connect6.CtrlConnectDDNet()
        elif unpacker.remaining_size() > 500: # TODO: this is a totally random number
            packet.set_version('0.6.5')
            msg = connect6.CtrlConnect65()
        else:
            raise ValueError(f"Error: invalid 0.6 control connect message data={unpacker.data().hex(sep = ' ')}")
    elif msg_id == twnet_parser.msg6.CTRL_CONNECT_ACCEPT:
        msg = accept_connect6.CtrlAcceptConnect()
    elif msg_id == twnet_parser.msg6.CTRL_ACCEPT:
        msg = ack_accept_connect6.CtrlAckAcceptConnect()
    elif msg_id == twnet_parser.msg6.CTRL_CLOSE:
        msg = close6.CtrlClose()

    if msg is None:
        raise ValueError(f"Error: unknown control message id={msg_id} data={unpacker.data().hex(sep = ' ')}")

    msg.unpack(unpacker, client)
    return msg



from twnet_parser.packer import Unpacker
from twnet_parser.snap_item import SnapItem
# from twnet_parser.item_matcher.match_snap7 import match_item

class Snapshot:
    def __init__(
            self,
            num_removed_items: int = 0,
            num_item_deltas: int = 0
    ) -> None:
        self.num_removed_items: int = num_removed_items
        self.num_item_deltas: int = num_item_deltas
        self.items: list[SnapItem] = []

    # expects the int compressed
    # data field of the snap message
    # not the whole snap message with crc
    # and the other fields
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.num_removed_items = unpacker.get_int()
        self.num_item_deltas = unpacker.get_int()
        unpacker.get_int() # unused by tw 0.7 NumTempItems

        # while unpacker.remaining_size() > 2:
        #     item = match_item(unpacker)
        #     if item is None:
        #         continue
        #     print(f"got item size={item.size} type={item.type_id}")

        #     item.unpack(unpacker)
        #     self.items.append(item)

        return True

    def pack(self) -> bytes:
        return b''




from typing import Protocol, Literal, Iterator, Any

from twnet_parser.chunk_header import ChunkHeader

class NetMessage(Protocol):
    message_type: Literal['system', 'game']
    message_name: str
    system_message: bool
    message_id: int
    header: ChunkHeader
    # TODO: what does the bool mean? document it!
    #       maybe even replace it with an int that indicates
    #       the amount of bytes consumed from data
    def unpack(self, data: bytes) -> bool:
        ...
    def pack(self) -> bytes:
        ...
    def __iter__(self) -> Iterator[tuple[str, Any]]:
        ...



from typing import Protocol, Literal, Iterator, Any

from twnet_parser.packer import Unpacker

class CtrlMessage(Protocol):
    message_type: Literal['control']
    message_name: str
    message_id: int
    def unpack(self, unpacker: Unpacker, we_are_a_client: bool = True) -> bool:
        ...
    def pack(self, we_are_a_client: bool = True) -> bytes:
        ...
    def __iter__(self) -> Iterator[tuple[str, Any]]:
        ...



from typing import Literal

from twnet_parser.net_message import NetMessage

from twnet_parser.msg_matcher.game7 import match_game7
from twnet_parser.msg_matcher.system7 import match_system7

from twnet_parser.msg_matcher.game6 import match_game6
from twnet_parser.msg_matcher.system6 import match_system6

# could also be named ChunkParser
class MessageParser():
    # the first byte of data has to be the
    # first byte of a message PAYLOAD
    # NOT the whole packet with packet header
    # and NOT the whole message with chunk header
    def parse_game_message(
            self,
            version: Literal['0.6', '0.6.4', '0.6.5', '0.7'],
            msg_id: int,
            data: bytes
    ) -> NetMessage:
        if version == '0.6':
            return match_game6(msg_id, data)
        return match_game7(msg_id, data)
    def parse_sys_message(
            self,
            version: Literal['0.6', '0.6.4', '0.6.5', '0.7'],
            msg_id: int,
            data: bytes
    ) -> NetMessage:
        if version == '0.6':
            return match_system6(msg_id, data)
        return match_system7(msg_id, data)



# generated by scripts/generate_messages.py
from typing import Optional

import twnet_parser.msg6
from twnet_parser.net_message import NetMessage
from twnet_parser.ddnet_uuid import MsgDDNetUuid

import twnet_parser.messages6.game.sv_motd as \
       game6_sv_motd
import twnet_parser.messages6.game.sv_broadcast as \
       game6_sv_broadcast
import twnet_parser.messages6.game.sv_chat as \
       game6_sv_chat
import twnet_parser.messages6.game.sv_kill_msg as \
       game6_sv_kill_msg
import twnet_parser.messages6.game.sv_sound_global as \
       game6_sv_sound_global
import twnet_parser.messages6.game.sv_tune_params as \
       game6_sv_tune_params
import twnet_parser.messages6.game.sv_extra_projectile as \
       game6_sv_extra_projectile
import twnet_parser.messages6.game.sv_ready_to_enter as \
       game6_sv_ready_to_enter
import twnet_parser.messages6.game.sv_weapon_pickup as \
       game6_sv_weapon_pickup
import twnet_parser.messages6.game.sv_emoticon as \
       game6_sv_emoticon
import twnet_parser.messages6.game.sv_vote_clear_options as \
       game6_sv_vote_clear_options
import twnet_parser.messages6.game.sv_vote_option_list_add as \
       game6_sv_vote_option_list_add
import twnet_parser.messages6.game.sv_vote_option_add as \
       game6_sv_vote_option_add
import twnet_parser.messages6.game.sv_vote_option_remove as \
       game6_sv_vote_option_remove
import twnet_parser.messages6.game.sv_vote_set as \
       game6_sv_vote_set
import twnet_parser.messages6.game.sv_vote_status as \
       game6_sv_vote_status
import twnet_parser.messages6.game.cl_say as \
       game6_cl_say
import twnet_parser.messages6.game.cl_set_team as \
       game6_cl_set_team
import twnet_parser.messages6.game.cl_set_spectator_mode as \
       game6_cl_set_spectator_mode
import twnet_parser.messages6.game.cl_start_info as \
       game6_cl_start_info
import twnet_parser.messages6.game.cl_change_info as \
       game6_cl_change_info
import twnet_parser.messages6.game.cl_kill as \
       game6_cl_kill
import twnet_parser.messages6.game.cl_emoticon as \
       game6_cl_emoticon
import twnet_parser.messages6.game.cl_vote as \
       game6_cl_vote
import twnet_parser.messages6.game.cl_call_vote as \
       game6_cl_call_vote

def match_game6(msg_id: int, data: bytes) -> NetMessage:
    msg: Optional[NetMessage] = None

    if msg_id == twnet_parser.msg6.SV_MOTD:
        msg = game6_sv_motd.MsgSvMotd()
    elif msg_id == twnet_parser.msg6.SV_BROADCAST:
        msg = game6_sv_broadcast.MsgSvBroadcast()
    elif msg_id == twnet_parser.msg6.SV_CHAT:
        msg = game6_sv_chat.MsgSvChat()
    elif msg_id == twnet_parser.msg6.SV_KILL_MSG:
        msg = game6_sv_kill_msg.MsgSvKillMsg()
    elif msg_id == twnet_parser.msg6.SV_SOUND_GLOBAL:
        msg = game6_sv_sound_global.MsgSvSoundGlobal()
    elif msg_id == twnet_parser.msg6.SV_TUNE_PARAMS:
        msg = game6_sv_tune_params.MsgSvTuneParams()
    elif msg_id == twnet_parser.msg6.SV_EXTRA_PROJECTILE:
        msg = game6_sv_extra_projectile.MsgSvExtraProjectile()
    elif msg_id == twnet_parser.msg6.SV_READY_TO_ENTER:
        msg = game6_sv_ready_to_enter.MsgSvReadyToEnter()
    elif msg_id == twnet_parser.msg6.SV_WEAPON_PICKUP:
        msg = game6_sv_weapon_pickup.MsgSvWeaponPickup()
    elif msg_id == twnet_parser.msg6.SV_EMOTICON:
        msg = game6_sv_emoticon.MsgSvEmoticon()
    elif msg_id == twnet_parser.msg6.SV_VOTE_CLEAR_OPTIONS:
        msg = game6_sv_vote_clear_options.MsgSvVoteClearOptions()
    elif msg_id == twnet_parser.msg6.SV_VOTE_OPTION_LIST_ADD:
        msg = game6_sv_vote_option_list_add.MsgSvVoteOptionListAdd()
    elif msg_id == twnet_parser.msg6.SV_VOTE_OPTION_ADD:
        msg = game6_sv_vote_option_add.MsgSvVoteOptionAdd()
    elif msg_id == twnet_parser.msg6.SV_VOTE_OPTION_REMOVE:
        msg = game6_sv_vote_option_remove.MsgSvVoteOptionRemove()
    elif msg_id == twnet_parser.msg6.SV_VOTE_SET:
        msg = game6_sv_vote_set.MsgSvVoteSet()
    elif msg_id == twnet_parser.msg6.SV_VOTE_STATUS:
        msg = game6_sv_vote_status.MsgSvVoteStatus()
    elif msg_id == twnet_parser.msg6.CL_SAY:
        msg = game6_cl_say.MsgClSay()
    elif msg_id == twnet_parser.msg6.CL_SET_TEAM:
        msg = game6_cl_set_team.MsgClSetTeam()
    elif msg_id == twnet_parser.msg6.CL_SET_SPECTATOR_MODE:
        msg = game6_cl_set_spectator_mode.MsgClSetSpectatorMode()
    elif msg_id == twnet_parser.msg6.CL_START_INFO:
        msg = game6_cl_start_info.MsgClStartInfo()
    elif msg_id == twnet_parser.msg6.CL_CHANGE_INFO:
        msg = game6_cl_change_info.MsgClChangeInfo()
    elif msg_id == twnet_parser.msg6.CL_KILL:
        msg = game6_cl_kill.MsgClKill()
    elif msg_id == twnet_parser.msg6.CL_EMOTICON:
        msg = game6_cl_emoticon.MsgClEmoticon()
    elif msg_id == twnet_parser.msg6.CL_VOTE:
        msg = game6_cl_vote.MsgClVote()
    elif msg_id == twnet_parser.msg6.CL_CALL_VOTE:
        msg = game6_cl_call_vote.MsgClCallVote()

    # msg_id 0 is used by ddnet uuid messages
    # do not crash on those
    if msg is None:
        if msg_id != 0:
            raise ValueError(f"Error: unknown game message id={msg_id} data={data[0]}")
        msg = MsgDDNetUuid()

    msg.unpack(data)
    return msg



# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import SANITIZE_CC, pack_int, pack_str
from typing import Literal

class MsgInfo(PrettyPrint):
    def __init__(
            self,
            token: int = 0,
            version: str = 'default',
            name: str = 'default',
            hostname: str = 'default',
            map: str = 'default',
            game_type: str = 'default',
            flags: int = 0,
            skill_level: int = 0,
            num_players: int = 0,
            max_players: int = 0,
            num_clients: int = 0,
            max_clients: int = 0,
            clients: bytes = b''
    ) -> None:
        self.message_type: Literal['connless'] = 'connless'
        self.message_name: str = 'connless.info'
        self.message_id: list[int] = [255, 255, 255, 255, 105, 110, 102, 51]

        self.token: int = token
        self.version: str = version
        self.name: str = name
        self.hostname: str = hostname
        self.map: str = map
        self.game_type: str = game_type
        self.flags: int = flags
        self.skill_level: int = skill_level
        self.num_players: int = num_players
        self.max_players: int = max_players
        self.num_clients: int = num_clients
        self.max_clients: int = max_clients
        self.clients: bytes = clients

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

        yield 'token', self.token
        yield 'version', self.version
        yield 'name', self.name
        yield 'hostname', self.hostname
        yield 'map', self.map
        yield 'game_type', self.game_type
        yield 'flags', self.flags
        yield 'skill_level', self.skill_level
        yield 'num_players', self.num_players
        yield 'max_players', self.max_players
        yield 'num_clients', self.num_clients
        yield 'max_clients', self.max_clients
        yield 'clients', self.clients

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.token = unpacker.get_int()
        self.version = unpacker.get_str(SANITIZE_CC)
        self.name = unpacker.get_str(SANITIZE_CC)
        self.hostname = unpacker.get_str(SANITIZE_CC)
        self.map = unpacker.get_str(SANITIZE_CC)
        self.game_type = unpacker.get_str(SANITIZE_CC)
        self.flags = unpacker.get_int()
        self.skill_level = unpacker.get_int()
        self.num_players = unpacker.get_int()
        self.max_players = unpacker.get_int()
        self.num_clients = unpacker.get_int()
        self.max_clients = unpacker.get_int()
        self.clients = unpacker.get_raw()
        return True

    def pack(self) -> bytes:
        return pack_int(self.token) + \
            pack_str(self.version) + \
            pack_str(self.name) + \
            pack_str(self.hostname) + \
            pack_str(self.map) + \
            pack_str(self.game_type) + \
            pack_int(self.flags) + \
            pack_int(self.skill_level) + \
            pack_int(self.num_players) + \
            pack_int(self.max_players) + \
            pack_int(self.num_clients) + \
            pack_int(self.max_clients) + \
            self.clients


# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from typing import Literal

class MsgRequestCount(PrettyPrint):
    def __init__(
            self
    ) -> None:
        self.message_type: Literal['connless'] = 'connless'
        self.message_name: str = 'connless.request_count'
        self.message_id: list[int] = [255, 255, 255, 255, 99, 111, 117, 50]


    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        return True

    def pack(self) -> bytes:
        return b''


# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjPlayerInput(PrettyPrint):
    def __init__(
            self,
            direction: int = 0,
            target_x: int = 0,
            target_y: int = 0,
            jump: int = 0,
            fire: int = 0,
            hook: int = 0,
            player_flags: int = 0,
            wanted_weapon: int = 0,
            next_weapon: int = 0,
            prev_weapon: int = 0
    ) -> None:
        self.item_name: str = 'item.player_input'
        self.type_id: int = 1
        self.id: int = 0
        self.size: int = 10

        self.direction: int = direction
        self.target_x: int = target_x
        self.target_y: int = target_y
        self.jump: int = jump
        self.fire: int = fire
        self.hook: int = hook
        self.player_flags: int = player_flags
        self.wanted_weapon: int = wanted_weapon
        self.next_weapon: int = next_weapon
        self.prev_weapon: int = prev_weapon

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'direction', self.direction
        yield 'target_x', self.target_x
        yield 'target_y', self.target_y
        yield 'jump', self.jump
        yield 'fire', self.fire
        yield 'hook', self.hook
        yield 'player_flags', self.player_flags
        yield 'wanted_weapon', self.wanted_weapon
        yield 'next_weapon', self.next_weapon
        yield 'prev_weapon', self.prev_weapon

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'direction': self.direction,
            'target_x': self.target_x,
            'target_y': self.target_y,
            'jump': self.jump,
            'fire': self.fire,
            'hook': self.hook,
            'player_flags': self.player_flags,
            'wanted_weapon': self.wanted_weapon,
            'next_weapon': self.next_weapon,
            'prev_weapon': self.prev_weapon
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.direction = unpacker.get_int()
        self.target_x = unpacker.get_int()
        self.target_y = unpacker.get_int()
        self.jump = unpacker.get_int()
        self.fire = unpacker.get_int()
        self.hook = unpacker.get_int()
        self.player_flags = unpacker.get_int()
        self.wanted_weapon = unpacker.get_int()
        self.next_weapon = unpacker.get_int()
        self.prev_weapon = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.direction) + \
            pack_int(self.target_x) + \
            pack_int(self.target_y) + \
            pack_int(self.jump) + \
            pack_int(self.fire) + \
            pack_int(self.hook) + \
            pack_int(self.player_flags) + \
            pack_int(self.wanted_weapon) + \
            pack_int(self.next_weapon) + \
            pack_int(self.prev_weapon)


#!/usr/bin/env python

from typing import Union, cast, Literal

from twnet_parser.error import ParserError
from twnet_parser.packer import Unpacker
from twnet_parser.message_parser import MessageParser
from twnet_parser.net_message import NetMessage
from twnet_parser.ctrl_message import CtrlMessage
from twnet_parser.connless_message import ConnlessMessage
from twnet_parser.chunk_header import ChunkHeader, ChunkFlags
from twnet_parser.msg_matcher.control7 import match_control7
from twnet_parser.msg_matcher.connless7 import match_connless7
from twnet_parser.msg_matcher.control6 import match_control6
from twnet_parser.msg_matcher.connless6 import match_connless6

import twnet_parser.huffman
from twnet_parser.packet6 import CONNLESS_PACKET_HEADER6_SIZE, TwPacket6, PacketHeaderParser6
from twnet_parser.packet7 import CONNLESS_PACKET_HEADER7_SIZE
from twnet_parser.packet7 import CHUNKFLAG7_RESEND, CHUNKFLAG7_VITAL, PACKET_HEADER7_SIZE
from twnet_parser.packet7 import PacketHeaderParser7, TwPacket7
import twnet_parser.serialize

class ChunkHeaderParser:
    def parse_flags7(self, data: bytes) -> ChunkFlags:
        # FFss ssss  xxss ssss
        flag_bits = (data[0] >> 6) & 0x03
        flags = ChunkFlags()
        flags.resend = (flag_bits & CHUNKFLAG7_RESEND) != 0
        flags.vital = (flag_bits & CHUNKFLAG7_VITAL) != 0
        return flags

    # the first byte of data has to be the
    # first byte of the chunk header
    def parse_header7(self, data: bytes) -> ChunkHeader:
        header = ChunkHeader(version = '0.7')
        header.flags = self.parse_flags7(data)
        # ffSS SSSS  xxSS SSSS
        header.size = ((data[0] & 0x3F) << 6) | (data[1] & 0x3F)
        if header.flags.vital:
            # ffss ssss  XXss ssss
            header.seq = ((data[1] & 0xC0) << 2) | data[2]
        return header

    # the first byte of data has to be the
    # first byte of the chunk header
    def parse_header6(self, data: bytes) -> ChunkHeader:
        header = ChunkHeader(version = '0.6')
        header.flags = self.parse_flags7(data)
        header.size = ((data[0] & 0x3F) << 4) | (data[1] & 0xF)
        if header.flags.vital:
            header.seq = ((data[1] & 0xF0) << 2) | data[2]
        return header

class PacketParser():
    def __init__(self, ignore_errors = False) -> None:
        self.version: Literal['0.6', '0.6.4', '0.6.5', '0.7'] = '0.7'
        # by default we throw on unexpected values
        self.ignore_errors = ignore_errors

    # the first byte of data has to be the
    # first byte of a message chunk
    # NOT the whole packet with packet header
    def get_messages(self, data: bytes, packet: Union[TwPacket6, TwPacket7]) -> list[NetMessage]:
        messages: list[NetMessage] = []
        i = 0
        while i < len(data) and len(messages) < (packet.header.num_chunks or 0):
            msg = self.get_message(data[i:])
            if msg.header.size is None:
                raise ValueError('header size is not set')
            i += msg.header.size + 2 # header + msg id = 3
            if msg.header.flags.vital:
                i += 1
            messages.append(msg)
        if i < len(data):
            num_bytes: int = len(data) - i
            if num_bytes == 4 and packet.version == '0.6':
                # if there are exactly 4 bytes left over at the end
                # of the message we assume it is the ddnet TKEN extension
                packet.is_ddnet = True
                packet.header.token = data[i:]
            elif not self.ignore_errors:
                raise ParserError(
                    f"Got {len(messages)}/{packet.header.num_chunks} messages " +
                    f"but there are still {num_bytes} bytes remaining"
                )
        return messages

    # the first byte of data has to be the
    # first byte of a message chunk
    # NOT the whole packet with packet header
    def get_message(self, data: bytes) -> NetMessage:
        if self.version == '0.6':
            chunk_header = ChunkHeaderParser().parse_header6(data)
        else:
            chunk_header = ChunkHeaderParser().parse_header7(data)
        i = 2
        if chunk_header.flags.vital:
            i += 1
        unpacker = Unpacker(data[i:])
        msg_id: int = unpacker.get_int()
        i += 1
        sys: bool = (msg_id & 1) == 1
        msg_id >>= 1
        msg: NetMessage
        if sys:
            msg = MessageParser().parse_sys_message(self.version, msg_id, unpacker.get_raw())
        else:
            msg = MessageParser().parse_game_message(self.version, msg_id, unpacker.get_raw())
        msg.header = chunk_header
        return msg

    def parse6(self, data: bytes, client: bool) -> TwPacket6:
        pck = TwPacket6(version = '0.6')
        # can be overwritten with 0.6.4 or 0.6.5 if we detect something specific
        self.version = '0.6'
        # TODO: what is the most performant way in python to do this?
        #       heap allocating a PacketHeaderParser7 just to bundle a bunch of
        #       methods that do not share state seems like a waste of performance
        #       would this be nicer with class methods?
        pck.header = PacketHeaderParser6().parse_header(data)
        # TODO: this is wrong!
        #       0.6.4 headers are 3 bytes
        #       and in 0.6.5 only if the token flag is set they are 7 bytes
        header_size = PACKET_HEADER7_SIZE
        if pck.header.flags.connless:
            header_size = CONNLESS_PACKET_HEADER6_SIZE
        elif not pck.header.flags.token:
            header_size = 3
        pck.payload_raw = data[header_size:]
        pck.payload_decompressed = pck.payload_raw
        if pck.header.flags.control:
            unpacker = Unpacker(data[header_size+1:])
            ctrl_msg: CtrlMessage = match_control6(data[header_size], unpacker, pck, client)
            pck.messages.append(ctrl_msg)
            if unpacker.remaining_size() == 4:
                # if there are exactly 4 bytes left over at the end
                # of the packet we assume it is the ddnet TKEN extension
                pck.is_ddnet = True
                pck.header.token = unpacker.get_raw(4)
                pck.set_version('0.6.4') # ddnet is always 0.6.4
            elif not self.ignore_errors and unpacker.remaining_size() > 0:
                raise ParserError(
                    f"Got control message with {unpacker.remaining_size()} "
                     "unexpected bytes at the end"
                )
            return pck
        if pck.header.flags.connless:
            connless_msg: ConnlessMessage = match_connless6(data[header_size:14], data[14:])
            pck.messages.append(connless_msg)
            return pck
        if pck.header.flags.compression:
            payload = pck.payload_raw
            pck.payload_decompressed = twnet_parser.huffman.decompress(payload)
        pck.messages = cast(
                list[Union[CtrlMessage, NetMessage, ConnlessMessage]],
                self.get_messages(pck.payload_decompressed, pck))
        return pck

    def parse7(self, data: bytes, client: bool) -> TwPacket7:
        pck = TwPacket7(version = '0.7')
        self.version = '0.7'
        # TODO: what is the most performant way in python to do this?
        #       heap allocating a PacketHeaderParser7 just to bundle a bunch of
        #       methods that do not share state seems like a waste of performance
        #       would this be nicer with class methods?
        pck.header = PacketHeaderParser7().parse_header(data)
        header_size = PACKET_HEADER7_SIZE
        if pck.header.flags.connless:
            header_size = CONNLESS_PACKET_HEADER7_SIZE
        pck.payload_raw = data[header_size:]
        pck.payload_decompressed = pck.payload_raw
        if pck.header.flags.control:
            ctrl_msg: CtrlMessage = match_control7(data[header_size], data[8:], client)
            pck.messages.append(ctrl_msg)
            return pck
        if pck.header.flags.connless:
            connless_msg: ConnlessMessage = match_connless7(data[header_size:17], data[17:])
            pck.messages.append(connless_msg)
            return pck
        if pck.header.flags.compression:
            payload = pck.payload_raw
            pck.payload_decompressed = twnet_parser.huffman.decompress(payload)
        pck.messages = cast(
                list[Union[CtrlMessage, NetMessage, ConnlessMessage]],
                self.get_messages(pck.payload_decompressed, pck))
        return pck

def parse6(data: bytes, ignore_errors: bool = False, we_are_a_client: bool = True) -> TwPacket6:
    return PacketParser(ignore_errors).parse6(data, we_are_a_client)

def parse7(data: bytes, ignore_errors: bool = False, we_are_a_client: bool = True) -> TwPacket7:
    return PacketParser(ignore_errors).parse7(data, we_are_a_client)



# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjCharacterCore(PrettyPrint):
    def __init__(
            self,
            tick: int = 0,
            x: int = 0,
            y: int = 0,
            vel_x: int = 0,
            vel_y: int = 0,
            angle: int = 0,
            direction: int = 0,
            jumped: int = 0,
            hooked_player: int = 0,
            hook_state: int = 0,
            hook_tick: int = 0,
            hook_x: int = 0,
            hook_y: int = 0,
            hook_dx: int = 0,
            hook_dy: int = 0
    ) -> None:
        self.item_name: str = 'item.character_core'
        self.type_id: int = 9
        self.id: int = 0
        self.size: int = 15

        self.tick: int = tick
        self.x: int = x
        self.y: int = y
        self.vel_x: int = vel_x
        self.vel_y: int = vel_y
        self.angle: int = angle
        self.direction: int = direction
        self.jumped: int = jumped
        self.hooked_player: int = hooked_player
        self.hook_state: int = hook_state
        self.hook_tick: int = hook_tick
        self.hook_x: int = hook_x
        self.hook_y: int = hook_y
        self.hook_dx: int = hook_dx
        self.hook_dy: int = hook_dy

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'tick', self.tick
        yield 'x', self.x
        yield 'y', self.y
        yield 'vel_x', self.vel_x
        yield 'vel_y', self.vel_y
        yield 'angle', self.angle
        yield 'direction', self.direction
        yield 'jumped', self.jumped
        yield 'hooked_player', self.hooked_player
        yield 'hook_state', self.hook_state
        yield 'hook_tick', self.hook_tick
        yield 'hook_x', self.hook_x
        yield 'hook_y', self.hook_y
        yield 'hook_dx', self.hook_dx
        yield 'hook_dy', self.hook_dy

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'tick': self.tick,
            'x': self.x,
            'y': self.y,
            'vel_x': self.vel_x,
            'vel_y': self.vel_y,
            'angle': self.angle,
            'direction': self.direction,
            'jumped': self.jumped,
            'hooked_player': self.hooked_player,
            'hook_state': self.hook_state,
            'hook_tick': self.hook_tick,
            'hook_x': self.hook_x,
            'hook_y': self.hook_y,
            'hook_dx': self.hook_dx,
            'hook_dy': self.hook_dy
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.tick = unpacker.get_int()
        self.x = unpacker.get_int()
        self.y = unpacker.get_int()
        self.vel_x = unpacker.get_int()
        self.vel_y = unpacker.get_int()
        self.angle = unpacker.get_int()
        self.direction = unpacker.get_int()
        self.jumped = unpacker.get_int()
        self.hooked_player = unpacker.get_int()
        self.hook_state = unpacker.get_int()
        self.hook_tick = unpacker.get_int()
        self.hook_x = unpacker.get_int()
        self.hook_y = unpacker.get_int()
        self.hook_dx = unpacker.get_int()
        self.hook_dy = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.tick) + \
            pack_int(self.x) + \
            pack_int(self.y) + \
            pack_int(self.vel_x) + \
            pack_int(self.vel_y) + \
            pack_int(self.angle) + \
            pack_int(self.direction) + \
            pack_int(self.jumped) + \
            pack_int(self.hooked_player) + \
            pack_int(self.hook_state) + \
            pack_int(self.hook_tick) + \
            pack_int(self.hook_x) + \
            pack_int(self.hook_y) + \
            pack_int(self.hook_dx) + \
            pack_int(self.hook_dy)


# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import pack_int
from typing import Literal, Optional

class MsgRconAuthStatus(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            auth_level: int = 0,
            receive_commands: int = 0
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'system'
        self.message_name: str = 'rcon_auth_status'
        self.system_message: bool = True
        self.message_id: int = 10
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.6')
        self.header: ChunkHeader = chunk_header

        self.auth_level: Optional[int] = auth_level
        self.receive_commands: Optional[int] = receive_commands

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'auth_level', self.auth_level
        yield 'receive_commands', self.receive_commands

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.auth_level = unpacker.get_optional_int()
        self.receive_commands = unpacker.get_optional_int()
        return True

    def pack(self) -> bytes:
        return (pack_int(self.auth_level) if self.auth_level is not None else b'') + \
            (pack_int(self.receive_commands) if self.receive_commands is not None else b'')


import hashlib

UUID_MAXSTRSIZE = 37
"""
12345678-0123-5678-0123-567890123456
"""

UUID_INVALID = -2

UUID_UNKNOWN = -1

OFFSET_UUID = 1 << 16
"""
OFFSET_UUID (65536)
"""


TEEWORLDS_NAMESPACE = b'\xe0\x5d\xda\xaa\xc4\xe6\x4c\xfb\xb6\x42\x5d\x48\xe8\x0c\x00\x29'
"""
e05ddaaa-c4e6-4cfb-b642-5d48e80c0029
"""

class DDNetUuid:
    def __init__(self, type_id: int, name: str, uuid: bytes) -> None:
        self.name = name
        self.type_id = type_id
        self.uuid = uuid

class UuidManager:
    def __init__(self) -> None:
        self.uuids: list[DDNetUuid] = []

    def get_index(self, type_id: int) -> int:
        return type_id - OFFSET_UUID

    def get_id(self, type_index: int) -> int:
        return type_index + OFFSET_UUID

    def register_name(self, type_id: int, name: str) -> None:
        """
        Every ddnet ex message has a uuid and a unique type_id which is an
        integer incremented starting from OFFSET_UUID (65536)

        What is sent over the network can also be a type_index which is
        the type_id minus the OFFSET_UUID
        """
        self.uuids.append(DDNetUuid(type_id, name, self.calculate_uuid(name)))

    def get_uuid_by_type(self, type_id: int) -> DDNetUuid:
        return self.uuids[type_id]

    def calculate_uuid(self, name: str) -> bytes:
        md5_hash = hashlib.md5()
        md5_hash.update(TEEWORLDS_NAMESPACE)
        md5_hash.update(name.encode('utf-8'))

        digest = bytearray(md5_hash.digest())

        # set version 3 (UUID is generated by MD5 hashing a namespace identifier and a name)
        digest[6] &= 0x0f
        digest[6] |= 0x30

        # set variant 1 (RFC 4122)
        digest[8] &= 0x3f
        digest[8] |= 0x80

        return bytes(digest)




# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjPickup(PrettyPrint):
    def __init__(
            self,
            x: int = 0,
            y: int = 0,
            type: int = 0,
            subtype: int = 0
    ) -> None:
        self.item_name: str = 'item.pickup'
        self.type_id: int = 4
        self.id: int = 0
        self.size: int = 4

        self.x: int = x
        self.y: int = y
        self.type: int = type
        self.subtype: int = subtype

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'x', self.x
        yield 'y', self.y
        yield 'type', self.type
        yield 'subtype', self.subtype

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'x': self.x,
            'y': self.y,
            'type': self.type,
            'subtype': self.subtype
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.x = unpacker.get_int()
        self.y = unpacker.get_int()
        self.type = unpacker.get_int()
        self.subtype = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.x) + \
            pack_int(self.y) + \
            pack_int(self.type) + \
            pack_int(self.subtype)


from typing import Optional

from twnet_parser.packer import Unpacker
from twnet_parser.snap_item import SnapItem

import twnet_parser.obj7

from twnet_parser.snap7.player_input import ObjPlayerInput
from twnet_parser.snap7.projectile import ObjProjectile
from twnet_parser.snap7.laser import ObjLaser
from twnet_parser.snap7.pickup import ObjPickup
from twnet_parser.snap7.flag import ObjFlag
from twnet_parser.snap7.game_data import ObjGameData
from twnet_parser.snap7.game_data_team import ObjGameDataTeam
from twnet_parser.snap7.game_data_flag import ObjGameDataFlag
from twnet_parser.snap7.character_core import ObjCharacterCore
from twnet_parser.snap7.character import ObjCharacter
from twnet_parser.snap7.player_info import ObjPlayerInfo
from twnet_parser.snap7.spectator_info import ObjSpectatorInfo
from twnet_parser.snap7.de_client_info import ObjDeClientInfo
from twnet_parser.snap7.de_game_info import ObjDeGameInfo
from twnet_parser.snap7.de_tune_params import ObjDeTuneParams
from twnet_parser.snap7.common import ObjCommon
from twnet_parser.snap7.explosion import ObjExplosion
from twnet_parser.snap7.spawn import ObjSpawn
from twnet_parser.snap7.hammer_hit import ObjHammerHit
from twnet_parser.snap7.death import ObjDeath
from twnet_parser.snap7.sound_world import ObjSoundWorld
from twnet_parser.snap7.damage import ObjDamage
from twnet_parser.snap7.player_info_race import ObjPlayerInfoRace
from twnet_parser.snap7.game_data_race import ObjGameDataRace

def match_item(unpacker: Unpacker) -> Optional[SnapItem]:
    item_type = unpacker.get_int()
    item: SnapItem
    if item_type == twnet_parser.obj7.PLAYER_INPUT:
        item = ObjPlayerInput()
    elif item_type == twnet_parser.obj7.PROJECTILE:
        item = ObjProjectile()
    elif item_type == twnet_parser.obj7.LASER:
        item = ObjLaser()
    elif item_type == twnet_parser.obj7.PICKUP:
        item = ObjPickup()
    elif item_type == twnet_parser.obj7.FLAG:
        item = ObjFlag()
    elif item_type == twnet_parser.obj7.GAME_DATA:
        item = ObjGameData()
    elif item_type == twnet_parser.obj7.GAME_DATA_TEAM:
        item = ObjGameDataTeam()
    elif item_type == twnet_parser.obj7.GAME_DATA_FLAG:
        item = ObjGameDataFlag()
    elif item_type == twnet_parser.obj7.CHARACTER_CORE:
        item = ObjCharacterCore()
    elif item_type == twnet_parser.obj7.CHARACTER:
        item = ObjCharacter()
    elif item_type == twnet_parser.obj7.PLAYER_INFO:
        item = ObjPlayerInfo()
    elif item_type == twnet_parser.obj7.SPECTATOR_INFO:
        item = ObjSpectatorInfo()
    elif item_type == twnet_parser.obj7.DE_CLIENT_INFO:
        item = ObjDeClientInfo()
    elif item_type == twnet_parser.obj7.DE_GAME_INFO:
        item = ObjDeGameInfo()
    elif item_type == twnet_parser.obj7.DE_TUNE_PARAMS:
        item = ObjDeTuneParams()
    elif item_type == twnet_parser.obj7.COMMON:
        item = ObjCommon()
    elif item_type == twnet_parser.obj7.EXPLOSION:
        item = ObjExplosion()
    elif item_type == twnet_parser.obj7.SPAWN:
        item = ObjSpawn()
    elif item_type == twnet_parser.obj7.HAMMERHIT:
        item = ObjHammerHit()
    elif item_type == twnet_parser.obj7.DEATH:
        item = ObjDeath()
    elif item_type == twnet_parser.obj7.SOUND_WORLD:
        item = ObjSoundWorld()
    elif item_type == twnet_parser.obj7.DAMAGE:
        item = ObjDamage()
    elif item_type == twnet_parser.obj7.PLAYER_INFO_RACE:
        unpacker.get_int() # pop size
        item = ObjPlayerInfoRace()
    elif item_type == twnet_parser.obj7.GAME_DATA_RACE:
        unpacker.get_int() # pop size
        item = ObjGameDataRace()
    else:
        item_id = unpacker.get_int()
        size = unpacker.get_int()
        print(f"Skipping unknown_type={item_type} id={item_id} size={size}")
        for _ in range(size):
            unpacker.get_int()
        return None
    return item




#!/usr/bin/env python3

import sys

import dpkt
import twnet_parser.packet

arg_pcap = None
arg_version = 7

def print_tw_packets(pcap):
    for _ts, buf in pcap:
        eth = dpkt.ethernet.Ethernet(buf)
        ip = eth.data
        if not isinstance(ip.data, dpkt.udp.UDP):
            continue
        udp_payload = ip.data.data
        try:
            if arg_version == 6:
                packet = twnet_parser.packet.parse6(udp_payload)
            else:
                packet = twnet_parser.packet.parse7(udp_payload)
        except:
            continue
        names = [msg.message_name for msg in packet.messages]
        print(f"[TW{arg_version}] {', '.join(names)}")

def usage():
    print(f'usage: {sys.argv[0]} <pcap file> [-6|-7]')
    print('options:')
    print(' -6  try to parse as teeworlds 0.6 packets')
    print(' -7  try to parse as teeworlds 0.7 packets (default)')

if len(sys.argv) < 2:
    usage()
    exit(1)

for arg in sys.argv[1:]:
    if arg == '-6':
        arg_version = 6
    elif arg == '-7':
        arg_version = 7
    elif arg_pcap is None:
        arg_pcap = arg

if arg_pcap is None:
    usage()
    exit(1)

with open(arg_pcap, 'rb') as f:
    pcap = dpkt.pcap.Reader(f)
    print_tw_packets(pcap)



from typing import Protocol, Literal, Annotated, Iterator, Any

class ConnlessMessage(Protocol):
    message_type: Literal['connless']
    message_name: str
    message_id: Annotated[list[int], 8]
    def unpack(self, data: bytes) -> bool:
        ...
    def pack(self) -> bytes:
        ...
    def __iter__(self) -> Iterator[tuple[str, Any]]:
        ...



# pylint: disable=duplicate-code
from enum import Enum

class Emote(Enum):
    NORMAL: int = 0
    PAIN: int = 1
    HAPPY: int = 2
    SURPRISE: int = 3
    ANGRY: int = 4
    BLINK: int = 5

class Powerup(Enum):
    HEALTH: int = 0
    ARMOR: int = 1
    WEAPON: int = 2
    NINJA: int = 3

class Emoticon(Enum):
    OOP: int = 0
    EXCLAMATION: int = 1
    HEARTS: int = 2
    DROP: int = 3
    DOTDOT: int = 4
    MUSIC: int = 5
    SORRY: int = 6
    GHOST: int = 7
    SUSHI: int = 8
    SPLATTEE: int = 9
    DEVILTEE: int = 10
    ZOMG: int = 11
    ZZZ: int = 12
    WTF: int = 13
    EYES: int = 14
    QUESTION: int = 15

class Weapon(Enum):
    HAMMER: int = 0
    PISTOL: int = 1
    SHOTGUN: int = 2
    GRENADE: int = 3
    RIFLE: int = 4
    NINJA: int = 5

class Team(Enum):
    SPECTATORS: int = -1
    RED: int = 0
    BLUE: int = 1

class Sound(Enum):
    GUN_FIRE: int = 0
    SHOTGUN_FIRE: int = 1
    GRENADE_FIRE: int = 2
    HAMMER_FIRE: int = 3
    HAMMER_HIT: int = 4
    NINJA_FIRE: int = 5
    GRENADE_EXPLODE: int = 6
    NINJA_HIT: int = 7
    RIFLE_FIRE: int = 8
    RIFLE_BOUNCE: int = 9
    WEAPON_SWITCH: int = 10
    PLAYER_PAIN_SHORT: int = 11
    PLAYER_PAIN_LONG: int = 12
    BODY_LAND: int = 13
    PLAYER_AIRJUMP: int = 14
    PLAYER_JUMP: int = 15
    PLAYER_DIE: int = 16
    PLAYER_SPAWN: int = 17
    PLAYER_SKID: int = 18
    TEE_CRY: int = 19
    HOOK_LOOP: int = 20
    HOOK_ATTACH_GROUND: int = 21
    HOOK_ATTACH_PLAYER: int = 22
    HOOK_NOATTACH: int = 23
    PICKUP_HEALTH: int = 24
    PICKUP_ARMOR: int = 25
    PICKUP_GRENADE: int = 26
    PICKUP_SHOTGUN: int = 27
    PICKUP_NINJA: int = 28
    WEAPON_SPAWN: int = 29
    WEAPON_NOAMMO: int = 30
    HIT: int = 31
    CHAT_SERVER: int = 32
    CHAT_CLIENT: int = 33
    CHAT_HIGHLIGHT: int = 34
    CTF_DROP: int = 35
    CTF_RETURN: int = 36
    CTF_GRAB_PL: int = 37
    CTF_GRAB_EN: int = 38
    CTF_CAPTURE: int = 39
    MENU: int = 40



from typing import Protocol
from twnet_parser.packer import Unpacker

class SnapItem(Protocol):
    item_name: str
    type_id: int
    id: int
    size: int
    def unpack(self, unpacker: Unpacker) -> bool:
        ...
    def pack(self) -> bytes:
        ...




from twnet_parser.huffman_twnet_parser import huffman

LIBTW2_HUFFMAN = None

try:
    import libtw2_huffman as LIBTW2_HUFFMAN # type: ignore
except ImportError:
    LIBTW2_HUFFMAN = None

def backend_name() -> str:
    if LIBTW2_HUFFMAN:
        return 'rust-libtw2'
    return 'python-twnet_parser'

def compress(data: bytes) -> bytes:
    if LIBTW2_HUFFMAN:
        return LIBTW2_HUFFMAN.compress(data)
    return huffman.decompress(data)

def decompress(data: bytes) -> bytes:
    if LIBTW2_HUFFMAN:
        return LIBTW2_HUFFMAN.decompress(data)
    return huffman.decompress(data)



#!/usr/bin/env python3
"""
Teeworlds huffman compression

Example usage::

from huffman import Huffman

huffman = Huffman()
decompressed = huffman.decompress(bytes([
        174, 149, 19, 92, 9, 87, 194,
        22, 177, 86, 220, 218, 34, 56,
        185, 18, 156, 168, 184, 1
])))
print(decompressed) # => b'hello world'
"""

import functools

FREQUENCY_TABLE = [
	1 << 30, 4545, 2657, 431, 1950, 919, 444, 482, 2244,
        617, 838, 542, 715, 1814, 304, 240, 754, 212, 647, 186,
	283, 131, 146, 166, 543, 164, 167, 136, 179, 859, 363, 113, 157, 154, 204, 108, 137, 180, 202, 176,
	872, 404, 168, 134, 151, 111, 113, 109, 120, 126, 129, 100, 41, 20, 16, 22, 18, 18, 17, 19,
	16, 37, 13, 21, 362, 166, 99, 78, 95, 88, 81, 70, 83, 284, 91, 187, 77, 68, 52, 68,
	59, 66, 61, 638, 71, 157, 50, 46, 69, 43, 11, 24, 13, 19, 10, 12, 12, 20, 14, 9,
	20, 20, 10, 10, 15, 15, 12, 12, 7, 19, 15, 14, 13, 18, 35, 19, 17, 14, 8, 5,
	15, 17, 9, 15, 14, 18, 8, 10, 2173, 134, 157, 68, 188, 60, 170, 60, 194, 62, 175, 71,
	148, 67, 167, 78, 211, 67, 156, 69, 1674, 90, 174, 53, 147, 89, 181, 51, 174, 63, 163, 80,
	167, 94, 128, 122, 223, 153, 218, 77, 200, 110, 190, 73, 174, 69, 145, 66, 277, 143, 141, 60,
	136, 53, 180, 57, 142, 57, 158, 61, 166, 112, 152, 92, 26, 22, 21, 28, 20, 26, 30, 21,
	32, 27, 20, 17, 23, 21, 30, 22, 22, 21, 27, 25, 17, 27, 23, 18, 39, 26, 15, 21,
	12, 18, 18, 27, 20, 18, 15, 19, 11, 17, 33, 12, 18, 15, 19, 18, 16, 26, 17, 18,
	9, 10, 25, 22, 22, 17, 20, 16, 6, 16, 15, 20, 14, 18, 24, 335, 1517]

HUFFMAN_EOF_SYMBOL = 256

HUFFMAN_MAX_SYMBOLS = HUFFMAN_EOF_SYMBOL + 1
HUFFMAN_MAX_NODES = HUFFMAN_MAX_SYMBOLS * 2 - 1

HUFFMAN_LUTBITS = 10
HUFFMAN_LUTSIZE = 1 << HUFFMAN_LUTBITS
HUFFMAN_LUTMASK = HUFFMAN_LUTSIZE - 1

class HuffmanConstructNode:
    """
    Internal class to build the huffman tree
    """
    def __init__(self, node_id: int, frequency: int) -> None:
        self.node_id = node_id
        self.frequency = frequency

def compare_nodes_by_frequency_desc(
        node1: HuffmanConstructNode,
        node2: HuffmanConstructNode
    ) -> int:
    """
    Comparison callback for stable sort
    """
    if node2.frequency < node1.frequency:
        return -1
    if node2.frequency > node1.frequency:
        return 1
    return 0

class Node():
    """
    Internal class to represent the huffman tree
    """
    def __init__(self, bits, num_bits, symbol):
        self.bits = bits
        self.num_bits = num_bits
        self.leafs = [0, 0]
        self.symbol = symbol

class Huffman:
    """
    Teeworlds huffman compression

    Example usage::

    from huffman import Huffman

    huffman = Huffman()
    decompressed = huffman.decompress(bytes([
            174, 149, 19, 92, 9, 87, 194,
            22, 177, 86, 220, 218, 34, 56,
            185, 18, 156, 168, 184, 1
    ])))
    print(decompressed) # => b'hello world'
    """
    def setbits_r(self, node: Node, bits: int, depth: int) -> None:
        """
        Internal method to initialize the huffman tree.
        """
        if node.leafs[1] != 0xFFFF:
            self.setbits_r(self.nodes[node.leafs[1]], bits|(1<<depth), depth+1)
        if node.leafs[0] != 0xFFFF:
            self.setbits_r(self.nodes[node.leafs[0]], bits, depth+1)

        if node.num_bits != 0:
            node.bits = bits
            node.num_bits = depth

    def construct_tree(self, frequencies: list[int]) -> None:
        """
        Internal method to initialize the huffman tree.
        """
        nodes_left: list[HuffmanConstructNode] = [
            HuffmanConstructNode(0, 0) for _ in range(0, HUFFMAN_MAX_SYMBOLS)
        ]

        num_nodes_left = HUFFMAN_MAX_SYMBOLS

        for i in range(0, HUFFMAN_MAX_SYMBOLS):
            self.nodes[i].num_bits = 0xFFFFFFFF
            self.nodes[i].symbol = i
            self.nodes[i].leafs[0] = 0xFFFF
            self.nodes[i].leafs[1] = 0xFFFF

            if i == HUFFMAN_EOF_SYMBOL:
                nodes_left[i].frequency = 1
            else:
                nodes_left[i].frequency = frequencies[i]

            nodes_left[i].node_id = i

        self.num_nodes = HUFFMAN_MAX_SYMBOLS

        while num_nodes_left > 1:
            nodes_left = sorted(
                nodes_left,
                key=functools.cmp_to_key(compare_nodes_by_frequency_desc)
            )

            self.nodes[self.num_nodes].num_bits = 0
            self.nodes[self.num_nodes].leafs[0] = nodes_left[num_nodes_left-1].node_id
            self.nodes[self.num_nodes].leafs[1] = nodes_left[num_nodes_left-2].node_id
            nodes_left[num_nodes_left-2].node_id = self.num_nodes
            nodes_left[num_nodes_left-2].frequency = \
                nodes_left[num_nodes_left-1].frequency + \
                nodes_left[num_nodes_left-2].frequency
            self.num_nodes += 1
            num_nodes_left -= 1
        self.start_node = self.nodes[self.num_nodes-1]
        self.setbits_r(self.start_node, 0, 0)

    def __init__(self) -> None:
        self.nodes: list[Node] = [Node(0, 0, 0) for _ in range(0, HUFFMAN_MAX_NODES)]
        self.decoded_lut: list[Node] = [Node(0, 0, 0) for _ in range(0, HUFFMAN_LUTSIZE)]
        self.num_nodes = 0
        self.start_node = Node(0, 0, 0)

        self.construct_tree(FREQUENCY_TABLE)

        for i in range(0, HUFFMAN_LUTSIZE):
            bits = i
            node = self.start_node
            k = 0
            while k < HUFFMAN_LUTBITS:
                node = self.nodes[node.leafs[bits & 1]]
                bits >>= 1

                if node.num_bits != 0:
                    self.decoded_lut[i] = node
                    break
                k += 1

            if k == HUFFMAN_LUTBITS:
                self.decoded_lut[i] = node

    def compress(self, data: bytes) -> bytes:
        """
        Compresses data using the teeworlds frequency table
        """
        src_index = 0
        end = len(data)
        bits = 0
        bitcount = 0
        dst = bytearray()

        if len(data) == 0:
            return b''

        symbol = data[src_index]
        src_index += 1

        while src_index < end:
            bits |= self.nodes[symbol].bits << bitcount
            bitcount += self.nodes[symbol].num_bits

            symbol = data[src_index]
            src_index += 1

            while bitcount >= 8:
                dst.append(bits & 0xFF)
                bits >>= 8
                bitcount -= 8

        bits |= self.nodes[symbol].bits << bitcount
        bitcount += self.nodes[symbol].num_bits
        while bitcount >= 8:
            dst.append(bits & 0xFF)
            bits >>= 8
            bitcount -= 8

        bits |= self.nodes[HUFFMAN_EOF_SYMBOL].bits << bitcount
        bitcount += self.nodes[HUFFMAN_EOF_SYMBOL].num_bits
        while bitcount >= 8:
            dst.append(bits & 0xFF)
            bits >>= 8
            bitcount -= 8
        dst.append(bits)
        return bytes(dst)

    def decompress(self, data: bytes) -> bytes:
        """
        Decompresses data using the teeworlds frequency table
        """
        src_index = 0
        size = len(data)
        dst = bytearray()
        bits = 0
        bitcount = 0
        eof = self.nodes[HUFFMAN_EOF_SYMBOL]
        node = Node(0, 0, 0)

        while True:
            found_node = False
            if bitcount >= HUFFMAN_LUTBITS:
                node = self.decoded_lut[bits & HUFFMAN_LUTMASK]
                found_node = True

            while bitcount < 24 and src_index < size:
                bits |= data[src_index] << bitcount
                src_index += 1
                bitcount += 8

            if not found_node:
                node = self.decoded_lut[bits & HUFFMAN_LUTMASK]

            if node.num_bits != 0:
                bits >>= node.num_bits
                bitcount -= node.num_bits
            else:
                bits >>= HUFFMAN_LUTBITS
                bitcount -= HUFFMAN_LUTBITS

                while True:
                    node = self.nodes[node.leafs[bits & 1]]
                    bitcount -= 1
                    bits >>= 1
                    if node.num_bits != 0:
                        break

                    # no more bits, decoding error
                    if bitcount == 0:
                        print("no more bits")
                        return dst
            if node == eof:
                break
            dst.append(node.symbol)
        return bytes(dst)

huffman = Huffman()



# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int
import twnet_parser.enum7 as enum7

class ObjCharacter(PrettyPrint):
    def __init__(
            self,
            health: int = 0,
            armor: int = 0,
            ammo_count: int = 0,
            weapon: int = enum7.Weapon.HAMMER.value,
            emote: int = enum7.Emote.NORMAL.value,
            attack_tick: int = 0,
            triggered_events: int = 0
    ) -> None:
        self.item_name: str = 'item.character'
        self.type_id: int = 10
        self.id: int = 0
        self.size: int = 22

        self.health: int = health
        self.armor: int = armor
        self.ammo_count: int = ammo_count
        self.weapon: int = weapon
        self.emote: int = emote
        self.attack_tick: int = attack_tick
        self.triggered_events: int = triggered_events

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'health', self.health
        yield 'armor', self.armor
        yield 'ammo_count', self.ammo_count
        yield 'weapon', self.weapon
        yield 'emote', self.emote
        yield 'attack_tick', self.attack_tick
        yield 'triggered_events', self.triggered_events

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'health': self.health,
            'armor': self.armor,
            'ammo_count': self.ammo_count,
            'weapon': self.weapon,
            'emote': self.emote,
            'attack_tick': self.attack_tick,
            'triggered_events': self.triggered_events
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        # character_core
        self.tick = unpacker.get_int()
        self.x = unpacker.get_int()
        self.y = unpacker.get_int()
        self.vel_x = unpacker.get_int()
        self.vel_y = unpacker.get_int()
        self.angle = unpacker.get_int()
        self.direction = unpacker.get_int()
        self.jumped = unpacker.get_int()
        self.hooked_player = unpacker.get_int()
        self.hook_state = unpacker.get_int()
        self.hook_tick = unpacker.get_int()
        self.hook_x = unpacker.get_int()
        self.hook_y = unpacker.get_int()
        self.hook_dx = unpacker.get_int()
        self.hook_dy = unpacker.get_int()

        # character
        self.health = unpacker.get_int()
        self.armor = unpacker.get_int()
        self.ammo_count = unpacker.get_int()
        self.weapon = unpacker.get_int() # enum WEAPON
        self.emote = unpacker.get_int() # enum EMOTE
        self.attack_tick = unpacker.get_int()
        self.triggered_events = unpacker.get_int() # TODO: this is a flag
        return True

    def pack(self) -> bytes:
        return pack_int(self.health) + \
            pack_int(self.armor) + \
            pack_int(self.ammo_count) + \
            pack_int(self.weapon) + \
            pack_int(self.emote) + \
            pack_int(self.attack_tick) + \
            pack_int(self.triggered_events)


# generated by scripts/generate_messages.py
from typing import Optional

import twnet_parser.msg7
from twnet_parser.net_message import NetMessage
from twnet_parser.ddnet_uuid import MsgDDNetUuid

import twnet_parser.messages7.system.info as \
       system7_info
import twnet_parser.messages7.system.map_change as \
       system7_map_change
import twnet_parser.messages7.system.map_data as \
       system7_map_data
import twnet_parser.messages7.system.server_info as \
       system7_server_info
import twnet_parser.messages7.system.con_ready as \
       system7_con_ready
import twnet_parser.messages7.system.snap as \
       system7_snap
import twnet_parser.messages7.system.snap_empty as \
       system7_snap_empty
import twnet_parser.messages7.system.snap_single as \
       system7_snap_single
import twnet_parser.messages7.system.input_timing as \
       system7_input_timing
import twnet_parser.messages7.system.rcon_auth_on as \
       system7_rcon_auth_on
import twnet_parser.messages7.system.rcon_auth_off as \
       system7_rcon_auth_off
import twnet_parser.messages7.system.rcon_line as \
       system7_rcon_line
import twnet_parser.messages7.system.rcon_cmd_add as \
       system7_rcon_cmd_add
import twnet_parser.messages7.system.rcon_cmd_rem as \
       system7_rcon_cmd_rem
import twnet_parser.messages7.system.ready as \
       system7_ready
import twnet_parser.messages7.system.enter_game as \
       system7_enter_game
import twnet_parser.messages7.system.input as \
       system7_input
import twnet_parser.messages7.system.rcon_cmd as \
       system7_rcon_cmd
import twnet_parser.messages7.system.rcon_auth as \
       system7_rcon_auth
import twnet_parser.messages7.system.request_map_data as \
       system7_request_map_data
import twnet_parser.messages7.system.ping as \
       system7_ping
import twnet_parser.messages7.system.ping_reply as \
       system7_ping_reply
import twnet_parser.messages7.system.maplist_entry_add as \
       system7_maplist_entry_add
import twnet_parser.messages7.system.maplist_entry_rem as \
       system7_maplist_entry_rem

def match_system7(msg_id: int, data: bytes) -> NetMessage:
    msg: Optional[NetMessage] = None

    if msg_id == twnet_parser.msg7.INFO:
        msg = system7_info.MsgInfo()
    elif msg_id == twnet_parser.msg7.MAP_CHANGE:
        msg = system7_map_change.MsgMapChange()
    elif msg_id == twnet_parser.msg7.MAP_DATA:
        msg = system7_map_data.MsgMapData()
    elif msg_id == twnet_parser.msg7.SERVER_INFO:
        msg = system7_server_info.MsgServerInfo()
    elif msg_id == twnet_parser.msg7.CON_READY:
        msg = system7_con_ready.MsgConReady()
    elif msg_id == twnet_parser.msg7.SNAP:
        msg = system7_snap.MsgSnap()
    elif msg_id == twnet_parser.msg7.SNAP_EMPTY:
        msg = system7_snap_empty.MsgSnapEmpty()
    elif msg_id == twnet_parser.msg7.SNAP_SINGLE:
        msg = system7_snap_single.MsgSnapSingle()
    elif msg_id == twnet_parser.msg7.INPUT_TIMING:
        msg = system7_input_timing.MsgInputTiming()
    elif msg_id == twnet_parser.msg7.RCON_AUTH_ON:
        msg = system7_rcon_auth_on.MsgRconAuthOn()
    elif msg_id == twnet_parser.msg7.RCON_AUTH_OFF:
        msg = system7_rcon_auth_off.MsgRconAuthOff()
    elif msg_id == twnet_parser.msg7.RCON_LINE:
        msg = system7_rcon_line.MsgRconLine()
    elif msg_id == twnet_parser.msg7.RCON_CMD_ADD:
        msg = system7_rcon_cmd_add.MsgRconCmdAdd()
    elif msg_id == twnet_parser.msg7.RCON_CMD_REM:
        msg = system7_rcon_cmd_rem.MsgRconCmdRem()
    elif msg_id == twnet_parser.msg7.READY:
        msg = system7_ready.MsgReady()
    elif msg_id == twnet_parser.msg7.ENTER_GAME:
        msg = system7_enter_game.MsgEnterGame()
    elif msg_id == twnet_parser.msg7.INPUT:
        msg = system7_input.MsgInput()
    elif msg_id == twnet_parser.msg7.RCON_CMD:
        msg = system7_rcon_cmd.MsgRconCmd()
    elif msg_id == twnet_parser.msg7.RCON_AUTH:
        msg = system7_rcon_auth.MsgRconAuth()
    elif msg_id == twnet_parser.msg7.REQUEST_MAP_DATA:
        msg = system7_request_map_data.MsgRequestMapData()
    elif msg_id == twnet_parser.msg7.PING:
        msg = system7_ping.MsgPing()
    elif msg_id == twnet_parser.msg7.PING_REPLY:
        msg = system7_ping_reply.MsgPingReply()
    elif msg_id == twnet_parser.msg7.MAPLIST_ENTRY_ADD:
        msg = system7_maplist_entry_add.MsgMaplistEntryAdd()
    elif msg_id == twnet_parser.msg7.MAPLIST_ENTRY_REM:
        msg = system7_maplist_entry_rem.MsgMaplistEntryRem()

    # msg_id 0 is used by ddnet uuid messages
    # do not crash on those
    if msg is None:
        if msg_id != 0:
            raise ValueError(f"Error: unknown system message id={msg_id} data={data[0]}")
        msg = MsgDDNetUuid()

    msg.unpack(data)
    return msg



# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from typing import Literal, Optional
from twnet_parser.snap6.projectile import ObjProjectile

class MsgSvExtraProjectile(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            projectile: Optional[ObjProjectile] = None
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'sv_extra_projectile'
        self.system_message: bool = False
        self.message_id: int = 7
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.6')
        self.header: ChunkHeader = chunk_header

        if not projectile:
            projectile = ObjProjectile()
        self.projectile: ObjProjectile = projectile

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'projectile', self.projectile.to_dict_payload_only()

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.projectile.unpack(unpacker)
        return True

    def pack(self) -> bytes:
        return self.projectile.pack()


# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import pack_int
from typing import Literal, Optional

class MsgSvGameInfo(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            game_flags: int = 0,
            score_limit: int = 0,
            time_limit: int = 0,
            match_num: int = 0,
            match_current: int = 0
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'sv_game_info'
        self.system_message: bool = False
        self.message_id: int = 19
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.7')
        self.header: ChunkHeader = chunk_header

        self.game_flags: int = game_flags
        self.score_limit: int = score_limit
        self.time_limit: int = time_limit
        self.match_num: int = match_num
        self.match_current: int = match_current

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'game_flags', self.game_flags
        yield 'score_limit', self.score_limit
        yield 'time_limit', self.time_limit
        yield 'match_num', self.match_num
        yield 'match_current', self.match_current

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.game_flags = unpacker.get_int() # TODO: this is a flag
        self.score_limit = unpacker.get_int()
        self.time_limit = unpacker.get_int()
        self.match_num = unpacker.get_int()
        self.match_current = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.game_flags) + \
            pack_int(self.score_limit) + \
            pack_int(self.time_limit) + \
            pack_int(self.match_num) + \
            pack_int(self.match_current)


# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import SANITIZE_CC, pack_int, pack_str
from typing import Literal, Optional
import twnet_parser.enum7 as enum7

class MsgSvChat(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            mode: int = enum7.Chat.NONE.value,
            client_id: int = 0,
            target_id: int = 0,
            message: str = 'default'
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'sv_chat'
        self.system_message: bool = False
        self.message_id: int = 3
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.7')
        self.header: ChunkHeader = chunk_header

        self.mode: int = mode
        self.client_id: int = client_id
        self.target_id: int = target_id
        self.message: str = message

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'mode', self.mode
        yield 'client_id', self.client_id
        yield 'target_id', self.target_id
        yield 'message', self.message

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.mode = unpacker.get_int() # enum CHAT
        self.client_id = unpacker.get_int()
        self.target_id = unpacker.get_int()
        self.message = unpacker.get_str(SANITIZE_CC)
        return True

    def pack(self) -> bytes:
        return pack_int(self.mode) + \
            pack_int(self.client_id) + \
            pack_int(self.target_id) + \
            pack_str(self.message)


#!/usr/bin/env python3

import socket
import os
import sys
import hashlib
from typing import Optional

from twnet_parser.packet import TwPacket, parse7
from twnet_parser.messages7.control.token import CtrlToken
from twnet_parser.messages7.control.connect import CtrlConnect
from twnet_parser.messages7.control.close import CtrlClose
from twnet_parser.messages7.system.info import MsgInfo
from twnet_parser.messages7.system.map_change import MsgMapChange
from twnet_parser.messages7.system.request_map_data import MsgRequestMapData

from twnet_parser.constants import NET_MAX_PACKETSIZE, NET_MAX_SEQUENCE

def progress(have, want):
    if os.name == 'nt':
        os.system('cls')
    else:
        os.system('clear')
    percent = int(100 * have / want)
    bar = '=' * int(percent / 2)
    bar = bar[0:-1] + '>'
    print(f"[{bar:50}] {have}/{want} {percent}%")
    print(outfile)

if len(sys.argv) < 3:
    print("usage: download.py HOST PORT")
    print("description:")
    print("  connects to given server and downloads the map")
    print("  the file will be stored in the current working directory")
    print("  then the client disconnects and quits the program")
    print("example:")
    print("  download.py localhost 8303")
    exit(1)

host = sys.argv[1]
port = int(sys.argv[2])
dest_srv = (host, port)

print(f"Connecting to {host}:{port} ...")

sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
sock.bind(('', 0))

outfile = None
downloaded_chunks = 0
downloaded_bytes = 0
sent_vital = 0
got_vital = 0
got_seqs = set()
map_info: Optional[MsgMapChange] = None
my_token = b'\xff\xaa\xbb\xee'
srv_token = b'\xff\xff\xff\xff'

def send_msg(messages):
    global srv_token
    global got_vital
    global sent_vital
    if not isinstance(messages, list):
        messages = [messages]
    packet = TwPacket()
    packet.header.token = srv_token
    for msg in messages:
        if hasattr(msg, 'header'):
            if msg.header.flags.vital:
                sent_vital += 1
            msg.header.seq = sent_vital
        packet.messages.append(msg)
    packet.header.ack = got_vital
    sock.sendto(packet.pack(), dest_srv)

# TODO: we should be able to set this
# ctrl_token = CtrlToken(we_are_a_client = True)
ctrl_token = CtrlToken()
ctrl_token.response_token = my_token
send_msg(ctrl_token)

def verify_checksum(filename, sha256):
    with open(filename, "rb") as f:
        bytes = f.read()
        readable_hash = hashlib.sha256(bytes).hexdigest()
        if readable_hash == sha256:
            print("OK checksum matches.")
            return
        else:
            print(f"ERROR expected={sha256} got={readable_hash}")
    exit(1)

while True:
    data, addr = sock.recvfrom(NET_MAX_PACKETSIZE)
    packet = parse7(data)

    for msg in packet.messages:
        if hasattr(msg, 'header'):
            if msg.header.flags.vital and not msg.header.flags.resend:
                got_vital = (got_vital + 1) % NET_MAX_SEQUENCE
            if msg.header.seq in got_seqs:
                continue
            got_seqs.add(msg.header.seq)

        if msg.message_name == 'token':
            srv_token = msg.response_token

            ctrl_connect = CtrlConnect()
            ctrl_connect.response_token = my_token
            send_msg(ctrl_connect)
        elif msg.message_name == 'accept':
            info = MsgInfo()
            info.header.flags.vital = True
            send_msg(info)
        elif msg.message_name == 'map_change':
            map_info = msg
            outfile = f"{map_info.name}_{map_info.sha256.hex()}.map"
            if os.path.isfile(outfile):
                print(f"Error: file already exists '{outfile}'")
                exit(1)
            req = MsgRequestMapData()
            req.header.flags.vital = True
            send_msg(req)
        elif msg.message_name == 'map_data':
            if not outfile:
                continue
            downloaded_chunks += 1
            downloaded_bytes += msg.header.size
            with open(outfile, "ab") as map_file:
                map_file.write(msg.data)
            progress(downloaded_bytes, map_info.size)
            if downloaded_bytes >= map_info.size:
                print(f"finished downloading {map_info.name}")
                verify_checksum(outfile, map_info.sha256.hex())
                send_msg(CtrlClose())
                exit(0)
            if downloaded_chunks % map_info.num_response_chunks_per_request == 0:
                req = MsgRequestMapData()
                req.header.flags.vital = True
                send_msg(req)
        elif msg.message_name == 'close':
            print("disconnected")
            exit(1)



PLAYER_INPUT = 1
PROJECTILE = 2
LASER = 3
PICKUP = 4
FLAG = 5
GAME_DATA = 6
GAME_DATA_TEAM = 7
GAME_DATA_FLAG = 8
CHARACTER_CORE = 9
CHARACTER = 10
PLAYER_INFO = 11
SPECTATOR_INFO = 12
DE_CLIENT_INFO = 13
DE_GAME_INFO = 14
DE_TUNE_PARAMS = 15
COMMON = 16
EXPLOSION = 17
SPAWN = 18
HAMMERHIT = 19
DEATH = 20
SOUND_WORLD = 21
DAMAGE = 22
PLAYER_INFO_RACE = 23
GAME_DATA_RACE = 24



# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import pack_int
from typing import Literal, Optional

class MsgMapData(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            last: int = 0,
            crc: int = 0,
            chunk: int = 0,
            data_size: Optional[int] = None,
            data: bytes = b'\x00'
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'system'
        self.message_name: str = 'map_data'
        self.system_message: bool = True
        self.message_id: int = 3
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.6')
        self.header: ChunkHeader = chunk_header

        self.last: int = last
        self.crc: int = crc
        self.chunk: int = chunk
        self.data_size: int = data_size if data_size else len(data)
        self.data: bytes = data

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'last', self.last
        yield 'crc', self.crc
        yield 'chunk', self.chunk
        yield 'data', self.data

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.last = unpacker.get_int()
        self.crc = unpacker.get_int()
        self.chunk = unpacker.get_int()
        self.data_size = unpacker.get_int()
        self.data = unpacker.get_raw(self.data_size)
        return True

    def pack(self) -> bytes:
        return pack_int(self.last) + \
            pack_int(self.crc) + \
            pack_int(self.chunk) + \
            pack_int(self.data_size) + \
            self.data


# connless
# pylint: disable=duplicate-code

CONNLESS_HEARTBEAT = b'\xff\xff\xff\xffbea2'
CONNLESS_REQUEST_LIST = b'\xff\xff\xff\xffreq2'
CONNLESS_LIST = b'\xff\xff\xff\xfflis2'
CONNLESS_REQUEST_COUNT = b'\xff\xff\xff\xffcou2'
CONNLESS_COUNT = b'\xff\xff\xff\xffsiz2'
CONNLESS_REQUEST_INFO = b'\xff\xff\xff\xffgie3'
CONNLESS_INFO = b'\xff\xff\xff\xffinf3'
CONNLESS_FORWARD_CHECK = b'\xff\xff\xff\xfffw??'
CONNLESS_FORWARD_RESPONSE = b'\xff\xff\xff\xfffw!!'
CONNLESS_FORWARD_OK = b'\xff\xff\xff\xfffwok'
CONNLESS_FORWARD_ERROR = b'\xff\xff\xff\xfffwer'

# control

CTRL_KEEPALIVE = 0
CTRL_CONNECT = 1
CTRL_CONNECT_ACCEPT = 2
CTRL_ACCEPT = 3 # got remove in 0.6.5
CTRL_CLOSE = 4

# system
INFO = 1
MAP_CHANGE = 2 # sent when client should switch map
MAP_DATA = 3 # map transfer, contains a chunk of the map file
CON_READY = 4 # connection is ready, client should send start info
SNAP = 5 # normal snapshot, multiple parts
SNAP_EMPTY = 6 # empty snapshot
SNAP_SINGLE = 7
SNAP_SMALL = 8
INPUT_TIMING = 9	# reports how off the input was
RCON_AUTH_STATUS = 10# result of the authentication
RCON_LINE = 11 # line that should be printed to the remote console
AUTH_CHALLANGE = 12
AUTH_RESULT = 13
READY = 14
ENTER_GAME = 15
INPUT = 16 # contains the inputdata from the client
RCON_CMD = 17
RCON_AUTH = 18
REQUEST_MAP_DATA = 19
AUTH_START = 20
AUTH_RESPONSE = 21
PING = 22
PING_REPLY = 23
ERROR = 24
RCON_CMD_ADD = 25
RCON_CMD_REMOVE = 26

# game
INVALID = 0
SV_MOTD = 1
SV_BROADCAST = 2
SV_CHAT = 3
SV_KILL_MSG = 4
SV_SOUND_GLOBAL = 5
SV_TUNE_PARAMS = 6
SV_EXTRA_PROJECTILE = 7
SV_READY_TO_ENTER = 8
SV_WEAPON_PICKUP = 9
SV_EMOTICON = 10
SV_VOTE_CLEAR_OPTIONS = 11
SV_VOTE_OPTION_LIST_ADD = 12
SV_VOTE_OPTION_ADD = 13
SV_VOTE_OPTION_REMOVE = 14
SV_VOTE_SET = 15
SV_VOTE_STATUS = 16
CL_SAY = 17
CL_SET_TEAM = 18
CL_SET_SPECTATOR_MODE = 19
CL_START_INFO = 20
CL_CHANGE_INFO = 21
CL_KILL = 22
CL_EMOTICON = 23
CL_VOTE = 24
CL_CALL_VOTE = 25
NUM_GAME_MESSAGES = 26




# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int
import twnet_parser.enum6 as enum6

class ObjPlayerInfo(PrettyPrint):
    def __init__(
            self,
            local: int = 0,
            client_id: int = 0,
            team: int = enum6.Team.SPECTATORS.value,
            score: int = 0,
            latency: int = 0
    ) -> None:
        self.item_name: str = 'item.player_info'
        self.type_id: int = 10
        self.id: int = 0
        self.size: int = 5

        self.local: int = local
        self.client_id: int = client_id
        self.team: int = team
        self.score: int = score
        self.latency: int = latency

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'local', self.local
        yield 'client_id', self.client_id
        yield 'team', self.team
        yield 'score', self.score
        yield 'latency', self.latency

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'local': self.local,
            'client_id': self.client_id,
            'team': self.team,
            'score': self.score,
            'latency': self.latency
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.local = unpacker.get_int()
        self.client_id = unpacker.get_int()
        self.team = unpacker.get_int() # enum TEAM
        self.score = unpacker.get_int()
        self.latency = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.local) + \
            pack_int(self.client_id) + \
            pack_int(self.team) + \
            pack_int(self.score) + \
            pack_int(self.latency)


from typing import Optional, Literal

from twnet_parser.pretty_print import PrettyPrint

class ChunkFlags(PrettyPrint):
    def __init__(self):
        self.resend = False
        self.vital = False

    def __repr__(self):
        return "<class: '" + str(self.__class__.__name__) + "'>: " + str(self.__dict__)

    def __iter__(self):
        flags = []
        if self.resend:
            flags.append('resend')
        if self.vital:
            flags.append('vital')
        return iter(flags)

# same fields for 0.6 and 0.7
# different bit layout tho
class ChunkHeader(PrettyPrint):
    def __init__(self, version: Literal['0.6', '0.7']) -> None:
        self.version = version
        self.flags: ChunkFlags = ChunkFlags()
        self.size: Optional[int] = None
        # TODO: should seq be a optional?
        #       so it can be None for non vital packages
        #       this could turn downstream users logic errors into
        #       crashes which would be easier to detect
        #
        #       Or is None annoying because it crashes
        #       and pollutes the code with error checking?
        #       Also the teeworlds code uses -1
        #       doing the same for someone who knows the codebase
        #       could also be nice
        #
        #       update: self.size uses optional now
        #               so it can be computed automatically
        #               if unset
        self.seq: int = -1

    def __iter__(self):
        yield 'version', self.version
        yield 'flags', list(self.flags)
        yield 'size', self.size
        yield 'seq', self.seq

    def pack(self) -> bytes:
        flags: int = 0
        if self.flags.resend:
            flags |= 2
        if self.flags.vital:
            flags |= 1
        if self.size is None:
            self.size = 0
        data: bytearray
        if self.version == '0.6':
            data = bytearray([ \
                    ((flags & 0x03) << 6) | \
                    ((self.size >> 4) & 0x3f),
                    self.size & 0xf
            ])
            if self.flags.vital:
                data[1] |= (self.seq >> 2) & 0xf0
                data += bytes([self.seq & 0xff])
        else: # 0.7
            data = bytearray([ \
                    ((flags & 0x03) << 6) | \
                    ((self.size >> 6) & 0x3f),
                    self.size & 0x3f
            ])
            if self.flags.vital:
                data[1] |= (self.seq >> 2) & 0xc0
                data += bytes([self.seq & 0xff])
        return bytes(data)

    def __repr__(self):
        return "<class: '" + str(self.__class__.__name__) + "'>: " + str(dict(self))




from typing import Literal
from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker

import twnet_parser.msg6

"""
CtrlConnect65

with security token from 0.6.5
not compatible with 0.6.4 or earlier
also not compatible with 0.7 or later
"""
class CtrlConnect65(PrettyPrint):
    def __init__(
            self,
            response_token: bytes = b'\xff\xff\xff\xff'
    ) -> None:
        self.message_type: Literal['control'] = 'control'
        self.message_name: str = 'connect'
        self.message_id: int = twnet_parser.msg6.CTRL_CONNECT

        self.response_token: bytes = response_token

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

        yield 'response_token', self.response_token

    def unpack(self, unpacker: Unpacker, we_are_a_client: bool = True) -> bool:
        # reflection attack protection in 0.6 is unused
        if len(unpacker.data()) < 512:
            return False
        data = unpacker.get_raw(512)
        self.response_token = data[4:8]
        return True

    def pack(self, we_are_a_client: bool = True) -> bytes:
        return bytes(4) + self.response_token + bytes(504)

class CtrlConnect64(PrettyPrint):
    def __init__(
            self
    ) -> None:
        self.message_type: Literal['control'] = 'control'
        self.message_name: str = 'connect'
        self.message_id: int = twnet_parser.msg6.CTRL_CONNECT

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

    def unpack(self, unpacker: Unpacker, we_are_a_client: bool = True) -> bool:
        return True

    def pack(self, we_are_a_client: bool = True) -> bytes:
        return b''

class CtrlConnectDDNet(PrettyPrint):
    def __init__(
            self
    ) -> None:
        self.message_type: Literal['control'] = 'control'
        self.message_name: str = 'connect'
        self.message_id: int = twnet_parser.msg6.CTRL_CONNECT

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

    def unpack(self, unpacker: Unpacker, we_are_a_client: bool = True) -> bool:
        # TODO: we could throw some errors here

        # reflection attack protection in 0.6 is unused
        if len(unpacker.data()) != 8:
            return False
        data = unpacker.get_raw(8)
        return data == b'TKEN\xff\xff\xff\xff'

    def pack(self, we_are_a_client: bool = True) -> bytes:
        return b'TKEN\xff\xff\xff\xff'



# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import pack_int
from typing import Literal, Optional

class MsgSvTuneParams(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            ground_control_speed: float = 10,
            ground_control_accel: float = 2,
            ground_friction: float = 0.5,
            ground_jump_impulse: float = 13.2,
            air_jump_impulse: float = 12,
            air_control_speed: float = 5,
            air_control_accel: float = 1.5,
            air_friction: float = 0.95,
            hook_length: float = 380,
            hook_fire_speed: float = 80,
            hook_drag_accel: float = 3,
            hook_drag_speed: float = 15,
            gravity: float = 0.5,
            velramp_start: float = 550,
            velramp_range: float = 2000,
            velramp_curvature: float = 1.4,
            gun_curvature: float = 1.25,
            gun_speed: float = 2200,
            gun_lifetime: float = 2,
            shotgun_curvature: float = 1.25,
            shotgun_speed: float = 2750,
            shotgun_speeddiff: float = 0.8,
            shotgun_lifetime: float = 0.2,
            grenade_curvature: float = 7,
            grenade_speed: float = 1000,
            grenade_lifetime: float = 2,
            laser_reach: float = 800,
            laser_bounce_delay: float = 150,
            laser_bounce_num: float = 1,
            laser_bounce_cost: float = 0,
            laser_damage: float = 0.0,
            player_collision: float = 1,
            player_hooking: float = 1
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'sv_tune_params'
        self.system_message: bool = False
        self.message_id: int = 6
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.6')
        self.header: ChunkHeader = chunk_header

        self.ground_control_speed: float = ground_control_speed
        self.ground_control_accel: float = ground_control_accel
        self.ground_friction: float = ground_friction
        self.ground_jump_impulse: float = ground_jump_impulse
        self.air_jump_impulse: float = air_jump_impulse
        self.air_control_speed: float = air_control_speed
        self.air_control_accel: float = air_control_accel
        self.air_friction: float = air_friction
        self.hook_length: float = hook_length
        self.hook_fire_speed: float = hook_fire_speed
        self.hook_drag_accel: float = hook_drag_accel
        self.hook_drag_speed: float = hook_drag_speed
        self.gravity: float = gravity
        self.velramp_start: float = velramp_start
        self.velramp_range: float = velramp_range
        self.velramp_curvature: float = velramp_curvature
        self.gun_curvature: float = gun_curvature
        self.gun_speed: float = gun_speed
        self.gun_lifetime: float = gun_lifetime
        self.shotgun_curvature: float = shotgun_curvature
        self.shotgun_speed: float = shotgun_speed
        self.shotgun_speeddiff: float = shotgun_speeddiff
        self.shotgun_lifetime: float = shotgun_lifetime
        self.grenade_curvature: float = grenade_curvature
        self.grenade_speed: float = grenade_speed
        self.grenade_lifetime: float = grenade_lifetime
        self.laser_reach: float = laser_reach
        self.laser_bounce_delay: float = laser_bounce_delay
        self.laser_bounce_num: float = laser_bounce_num
        self.laser_bounce_cost: float = laser_bounce_cost
        self.laser_damage: float = laser_damage
        self.player_collision: float = player_collision
        self.player_hooking: float = player_hooking

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'ground_control_speed', self.ground_control_speed
        yield 'ground_control_accel', self.ground_control_accel
        yield 'ground_friction', self.ground_friction
        yield 'ground_jump_impulse', self.ground_jump_impulse
        yield 'air_jump_impulse', self.air_jump_impulse
        yield 'air_control_speed', self.air_control_speed
        yield 'air_control_accel', self.air_control_accel
        yield 'air_friction', self.air_friction
        yield 'hook_length', self.hook_length
        yield 'hook_fire_speed', self.hook_fire_speed
        yield 'hook_drag_accel', self.hook_drag_accel
        yield 'hook_drag_speed', self.hook_drag_speed
        yield 'gravity', self.gravity
        yield 'velramp_start', self.velramp_start
        yield 'velramp_range', self.velramp_range
        yield 'velramp_curvature', self.velramp_curvature
        yield 'gun_curvature', self.gun_curvature
        yield 'gun_speed', self.gun_speed
        yield 'gun_lifetime', self.gun_lifetime
        yield 'shotgun_curvature', self.shotgun_curvature
        yield 'shotgun_speed', self.shotgun_speed
        yield 'shotgun_speeddiff', self.shotgun_speeddiff
        yield 'shotgun_lifetime', self.shotgun_lifetime
        yield 'grenade_curvature', self.grenade_curvature
        yield 'grenade_speed', self.grenade_speed
        yield 'grenade_lifetime', self.grenade_lifetime
        yield 'laser_reach', self.laser_reach
        yield 'laser_bounce_delay', self.laser_bounce_delay
        yield 'laser_bounce_num', self.laser_bounce_num
        yield 'laser_bounce_cost', self.laser_bounce_cost
        yield 'laser_damage', self.laser_damage
        yield 'player_collision', self.player_collision
        yield 'player_hooking', self.player_hooking

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.ground_control_speed = unpacker.get_int() / 100.0
        self.ground_control_accel = unpacker.get_int() / 100.0
        self.ground_friction = unpacker.get_int() / 100.0
        self.ground_jump_impulse = unpacker.get_int() / 100.0
        self.air_jump_impulse = unpacker.get_int() / 100.0
        self.air_control_speed = unpacker.get_int() / 100.0
        self.air_control_accel = unpacker.get_int() / 100.0
        self.air_friction = unpacker.get_int() / 100.0
        self.hook_length = unpacker.get_int() / 100.0
        self.hook_fire_speed = unpacker.get_int() / 100.0
        self.hook_drag_accel = unpacker.get_int() / 100.0
        self.hook_drag_speed = unpacker.get_int() / 100.0
        self.gravity = unpacker.get_int() / 100.0
        self.velramp_start = unpacker.get_int() / 100.0
        self.velramp_range = unpacker.get_int() / 100.0
        self.velramp_curvature = unpacker.get_int() / 100.0
        self.gun_curvature = unpacker.get_int() / 100.0
        self.gun_speed = unpacker.get_int() / 100.0
        self.gun_lifetime = unpacker.get_int() / 100.0
        self.shotgun_curvature = unpacker.get_int() / 100.0
        self.shotgun_speed = unpacker.get_int() / 100.0
        self.shotgun_speeddiff = unpacker.get_int() / 100.0
        self.shotgun_lifetime = unpacker.get_int() / 100.0
        self.grenade_curvature = unpacker.get_int() / 100.0
        self.grenade_speed = unpacker.get_int() / 100.0
        self.grenade_lifetime = unpacker.get_int() / 100.0
        self.laser_reach = unpacker.get_int() / 100.0
        self.laser_bounce_delay = unpacker.get_int() / 100.0
        self.laser_bounce_num = unpacker.get_int() / 100.0
        self.laser_bounce_cost = unpacker.get_int() / 100.0
        self.laser_damage = unpacker.get_int() / 100.0
        self.player_collision = unpacker.get_int() / 100.0
        self.player_hooking = unpacker.get_int() / 100.0
        return True

    def pack(self) -> bytes:
        return pack_int(int(self.ground_control_speed * 100.0)) + \
            pack_int(int(self.ground_control_accel * 100.0)) + \
            pack_int(int(self.ground_friction * 100.0)) + \
            pack_int(int(self.ground_jump_impulse * 100.0)) + \
            pack_int(int(self.air_jump_impulse * 100.0)) + \
            pack_int(int(self.air_control_speed * 100.0)) + \
            pack_int(int(self.air_control_accel * 100.0)) + \
            pack_int(int(self.air_friction * 100.0)) + \
            pack_int(int(self.hook_length * 100.0)) + \
            pack_int(int(self.hook_fire_speed * 100.0)) + \
            pack_int(int(self.hook_drag_accel * 100.0)) + \
            pack_int(int(self.hook_drag_speed * 100.0)) + \
            pack_int(int(self.gravity * 100.0)) + \
            pack_int(int(self.velramp_start * 100.0)) + \
            pack_int(int(self.velramp_range * 100.0)) + \
            pack_int(int(self.velramp_curvature * 100.0)) + \
            pack_int(int(self.gun_curvature * 100.0)) + \
            pack_int(int(self.gun_speed * 100.0)) + \
            pack_int(int(self.gun_lifetime * 100.0)) + \
            pack_int(int(self.shotgun_curvature * 100.0)) + \
            pack_int(int(self.shotgun_speed * 100.0)) + \
            pack_int(int(self.shotgun_speeddiff * 100.0)) + \
            pack_int(int(self.shotgun_lifetime * 100.0)) + \
            pack_int(int(self.grenade_curvature * 100.0)) + \
            pack_int(int(self.grenade_speed * 100.0)) + \
            pack_int(int(self.grenade_lifetime * 100.0)) + \
            pack_int(int(self.laser_reach * 100.0)) + \
            pack_int(int(self.laser_bounce_delay * 100.0)) + \
            pack_int(int(self.laser_bounce_num * 100.0)) + \
            pack_int(int(self.laser_bounce_cost * 100.0)) + \
            pack_int(int(self.laser_damage * 100.0)) + \
            pack_int(int(self.player_collision * 100.0)) + \
            pack_int(int(self.player_hooking * 100.0))


# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int
from typing import Literal

class MsgRequestInfo(PrettyPrint):
    def __init__(
            self,
            token: int = 0
    ) -> None:
        self.message_type: Literal['connless'] = 'connless'
        self.message_name: str = 'connless.request_info'
        self.message_id: list[int] = [255, 255, 255, 255, 103, 105, 101, 51]

        self.token: int = token

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

        yield 'token', self.token

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.token = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.token)


# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjDeGameInfo(PrettyPrint):
    def __init__(
            self,
            game_flags: int = 0,
            score_limit: int = 0,
            time_limit: int = 0,
            match_num: int = 0,
            match_current: int = 0
    ) -> None:
        self.item_name: str = 'item.de_game_info'
        self.type_id: int = 14
        self.id: int = 0
        self.size: int = 5

        self.game_flags: int = game_flags
        self.score_limit: int = score_limit
        self.time_limit: int = time_limit
        self.match_num: int = match_num
        self.match_current: int = match_current

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'game_flags', self.game_flags
        yield 'score_limit', self.score_limit
        yield 'time_limit', self.time_limit
        yield 'match_num', self.match_num
        yield 'match_current', self.match_current

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'game_flags': self.game_flags,
            'score_limit': self.score_limit,
            'time_limit': self.time_limit,
            'match_num': self.match_num,
            'match_current': self.match_current
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.game_flags = unpacker.get_int() # TODO: this is a flag
        self.score_limit = unpacker.get_int()
        self.time_limit = unpacker.get_int()
        self.match_num = unpacker.get_int()
        self.match_current = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.game_flags) + \
            pack_int(self.score_limit) + \
            pack_int(self.time_limit) + \
            pack_int(self.match_num) + \
            pack_int(self.match_current)


from typing import Literal
from twnet_parser.packer import Unpacker
from twnet_parser.pretty_print import PrettyPrint

class CtrlToken(PrettyPrint):
    def __init__(
            self,
            response_token: bytes = b'\xff\xff\xff\xff'
    ) -> None:
        self.message_type: Literal['control'] = 'control'
        self.message_name: str = 'token'
        self.message_id: int = 5

        self.response_token: bytes = response_token

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

        yield 'response_token', self.response_token

    def unpack(self, unpacker: Unpacker, we_are_a_client: bool = True) -> bool:
        if not we_are_a_client:
            # anti reflection attack
            if len(unpacker.data()) < 512:
                return False
        self.response_token = unpacker.data()[0:4]
        return True

    def pack(self, we_are_a_client: bool = True) -> bytes:
        if we_are_a_client:
            return self.response_token + bytes(508)
        return self.response_token



from typing import Optional

import twnet_parser.msg7
from twnet_parser.ctrl_message import CtrlMessage

import twnet_parser.messages7.control.keep_alive as keep_alive7
import twnet_parser.messages7.control.connect as connect7
import twnet_parser.messages7.control.accept as accept7
import twnet_parser.messages7.control.close as close7
import twnet_parser.messages7.control.token as token7
from twnet_parser.packer import Unpacker

def match_control7(msg_id: int, data: bytes, client: bool) -> CtrlMessage:
    msg: Optional[CtrlMessage] = None

    if msg_id == twnet_parser.msg7.CTRL_KEEPALIVE:
        msg = keep_alive7.CtrlKeepAlive()
    elif msg_id == twnet_parser.msg7.CTRL_CONNECT:
        msg = connect7.CtrlConnect()
    elif msg_id == twnet_parser.msg7.CTRL_ACCEPT:
        msg = accept7.CtrlAccept()
    elif msg_id == twnet_parser.msg7.CTRL_CLOSE:
        msg = close7.CtrlClose()
    elif msg_id == twnet_parser.msg7.CTRL_TOKEN:
        msg = token7.CtrlToken()

    if msg is None:
        raise ValueError(f"Error: unknown control message id={msg_id} data={data[0]}")

    unpacker = Unpacker(data)
    msg.unpack(unpacker, client)
    return msg



# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjDamageInd(PrettyPrint):
    def __init__(
            self,
            angle: int = 0
    ) -> None:
        self.item_name: str = 'item.damage_ind'
        self.type_id: int = 20
        self.id: int = 0
        self.size: int = 3

        self.angle: int = angle

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'angle', self.angle

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'angle': self.angle
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        # common
        self.x = unpacker.get_int()
        self.y = unpacker.get_int()

        # damage_ind
        self.angle = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.angle)


#!/usr/bin/env python

# pylint: disable=duplicate-code

from typing import Union, cast, Optional, Literal, Iterator, Any
import json

from twnet_parser.packer import pack_int
from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.net_message import NetMessage
from twnet_parser.ctrl_message import CtrlMessage
from twnet_parser.connless_message import ConnlessMessage
from twnet_parser.constants import NET_MAX_SEQUENCE, NET_PACKETVERSION

import twnet_parser.huffman
import twnet_parser.serialize

# TODO: what is a nice pythonic way of storing those?
#       also does some version:: namespace thing make sense?
PACKETFLAG7_CONTROL = 1
PACKETFLAG7_RESEND = 2
PACKETFLAG7_COMPRESSION = 4
PACKETFLAG7_CONNLESS = 8

CHUNKFLAG7_VITAL = 1
CHUNKFLAG7_RESEND = 2

PACKET_HEADER7_SIZE = 7
CONNLESS_PACKET_HEADER7_SIZE = 9
CONNLESS_PACKET_HEADER6_SIZE = 6

class PacketFlags7(PrettyPrint):
    def __init__(self) -> None:
        self.control: Optional[bool] = None
        self.resend: Optional[bool] = None
        self.compression: Optional[bool] = None
        self.connless: Optional[bool] = None

    def __iter__(self):
        flags = []
        if self.control:
            flags.append('control')
        if self.resend:
            flags.append('resend')
        if self.compression:
            flags.append('compression')
        if self.connless:
            flags.append('connless')
        return iter(flags)

    def __repr__(self):
        return "<class: '" + str(self.__class__.__name__) + "'>: " + str(list(self))

class PacketHeader7(PrettyPrint):
    def __init__(
            self,
            flags: Optional[PacketFlags7] = None,
            ack: int = 0,
            token: bytes = b'\xff\xff\xff\xff',
            num_chunks: Optional[int] = None
    ) -> None:
        """
        If num_chunks is not set it will count
        the messages it was given when
        the pack() method is called
        """
        if not flags:
            flags = PacketFlags7()
        self.flags: PacketFlags7 = flags
        self.ack: int = ack % NET_MAX_SEQUENCE
        self.token: bytes = token
        self.num_chunks: Optional[int] = num_chunks

        # connless only
        self.connless_version: int = NET_PACKETVERSION
        self.response_token: bytes = b'\xff\xff\xff\xff'

    def __iter__(self):
        yield 'flags', list(self.flags)
        yield 'ack', self.ack
        yield 'token', self.token
        yield 'num_chunks', self.num_chunks

        if self.flags.connless:
            yield 'connless_version', self.connless_version
            yield 'response_token', self.response_token

    def pack(self) -> bytes:
        """
        Generate 7 byte teeworlds 0.7 packet header
        based on the current instance variable
        values.

        The layout is as follows
        6bit flags, 2bit ack
        8bit ack
        8bit chunks
        32bit token

        ffffffaa
        aaaaaaaa
        NNNNNNNN
        TTTTTTTT
        TTTTTTTT
        TTTTTTTT
        TTTTTTTT
        """
        flags = 0
        if self.flags.control:
            flags |= PACKETFLAG7_CONTROL
        if self.flags.resend:
            flags |= PACKETFLAG7_RESEND
        if self.flags.compression:
            flags |= PACKETFLAG7_COMPRESSION
        if self.flags.connless:
            flags |= PACKETFLAG7_CONNLESS
        if self.num_chunks is None:
            self.num_chunks = 0
        if self.flags.connless:
            return bytes([ \
                ((PACKETFLAG7_CONNLESS<<2)&0xfc) | (self.connless_version&0x03)
            ]) + self.token + self.response_token
        return bytes([ \
            ((flags << 2)&0xfc) | ((self.ack>>8)&0x03), \
            self.ack&0xff, \
            self.num_chunks \
        ]) + self.token

class TwPacket7(PrettyPrint):
    def __init__(self, version: Literal['0.7'] = '0.7') -> None:
        # version is 0.7 at all times for TwPacket7
        self._version: Literal['0.7'] = version

        # There is no packet level ddnet protocol based on 0.7
        # There are still 0.7 snapshot and net message extensions
        # but no custom header/flags or tokens
        self.is_ddnet = False

        # packet payload as raw bytes
        # might be huffman compressed if header.flags.compression is set
        self.payload_raw: bytes = b''

        # packet payload after huffman decompression
        # can match payload_raw if no compression was applied
        self.payload_decompressed: bytes = b''

        # information about the packet
        # metadata that is not contained in the payload
        self.header: PacketHeader7
        if self.version == '0.7':
            self.header = PacketHeader7()
        else:
            raise ValueError(f"Error: invalid packet version '{self.version}'")
        self.messages: list[Union[CtrlMessage, NetMessage, ConnlessMessage]] = []

    def __iter__(self) -> Iterator[tuple[str, Any]]:
        yield 'version', self.version
        yield 'payload_raw', self.payload_raw
        yield 'payload_decompressed', self.payload_decompressed
        yield 'header', dict(self.header)

        yield 'messages', [dict(msg) for msg in self.messages]

    def to_json(self) -> str:
        return json.dumps(
            dict(self),
            indent=2,
            sort_keys=False,
            default=twnet_parser.serialize.bytes_to_hex
        )

    @property
    def version(self) -> Literal['0.7']:
        return self._version

    def pack(self, we_are_a_client = True) -> bytes:
        payload: bytes = b''
        msg: Union[CtrlMessage, NetMessage, ConnlessMessage]
        is_control: bool = False
        is_connless: bool = False
        for msg in self.messages:
            if msg.message_type == 'connless':
                is_connless = True
                msg = cast(ConnlessMessage, msg)
                payload += bytes(msg.message_id)
                payload += msg.pack()
            elif msg.message_type == 'control':
                is_control = True
                msg = cast(CtrlMessage, msg)
                payload += pack_int(msg.message_id)
                payload += msg.pack(we_are_a_client)
            else: # game or system message
                msg = cast(NetMessage, msg)
                msg_payload: bytes = pack_int(
                    (msg.message_id<<1) |
                    (int)(msg.system_message)
                )
                msg_payload += msg.pack()
                if msg.header.size is None:
                    msg.header.size = len(msg_payload)
                payload += msg.header.pack()
                payload += msg_payload
        if self.header.num_chunks is None:
            if is_control:
                self.header.num_chunks = 0
            else:
                self.header.num_chunks = len(self.messages)
        if is_control:
            if self.header.flags.control is None:
                self.header.flags.control = True
        if is_connless:
            if self.header.flags.connless is None:
                self.header.flags.connless = True
        if self.header.flags.compression:
            payload = twnet_parser.huffman.compress(payload)
        return self.header.pack() + payload

class PacketHeaderParser7():
    def parse_flags7(self, data: bytes) -> PacketFlags7:
        # FFFF FFaa
        flag_bits = (data[0] & 0xfc) >> 2
        flags = PacketFlags7()
        flags.control = (flag_bits & PACKETFLAG7_CONTROL) != 0
        flags.resend = (flag_bits & PACKETFLAG7_RESEND) != 0
        flags.compression = (flag_bits & PACKETFLAG7_COMPRESSION) != 0
        flags.connless = (flag_bits & PACKETFLAG7_CONNLESS) != 0
        return flags

    def parse_ack(self, header_bytes: bytes) -> int:
        # ffAA AAAA AAAA
        return ((header_bytes[0] & 0x3) << 8) | header_bytes[1]

    def parse_num_chunks(self, header_bytes: bytes) -> int:
        # TODO: not sure if this is correct
        return header_bytes[2]

    def parse_token(self, header_bytes: bytes) -> bytes:
        return header_bytes[3:7]

    def parse_header(self, data: bytes) -> PacketHeader7:
        header = PacketHeader7()
        # bits 2..5
        header.flags = self.parse_flags7(data)
        if header.flags.connless:
            # TODO: do not hardcode version field
            #       but actually read the bits
            header.connless_version = NET_PACKETVERSION
            header.token = data[1:5]
            header.response_token = data[5:9]
        else:
            # bits 6..16
            header.ack = self.parse_ack(data)
            # bits 17..25
            header.num_chunks = self.parse_num_chunks(data)
            # bits 16..57
            header.token = self.parse_token(data)
        return header




#!/usr/bin/env python

from typing import Final, Literal, Optional

from twnet_parser.master_server import MastersrvAddr

# Before chaning the current packer code to extend it
# Consider having two packers
#
# One being this one. That is simple and fast.
# Has no state or side effects.
# Does no error checking.
#
# And another one that sanitizes the data
# will throw errors unexpected data
# return the amount of read and written bytes
# and is attached to a class instance that
# keeps track of a state

NO_SANITIZE: Final[int] = 0
SANITIZE: Final[int]  = 1
SANITIZE_CC: Final[int] = 2
SKIP_START_WHITESPACES: Final[int] = 3

class Unpacker():
    def __init__(self, data: bytes) -> None:
        self._data = data
        self.idx = 0

    # first byte of the current buffer
    def byte(self) -> int:
        return self._data[self.idx]

    def data(self) -> bytes:
        return self._data[self.idx:]

    def remaining_size(self) -> int:
        return len(self._data) - self.idx

    def get_raw(self, size: int = -1) -> bytes:
        if size == -1:
            return self.data()
        end: int = self.idx + size
        data: bytes = self._data[self.idx:end]
        self.idx = end
        return data

    def get_packed_addresses(self) -> list[MastersrvAddr]:
        servers: list[MastersrvAddr] = []
        while len(self.data()) >= 18:
            ipaddr = self.get_raw(16)
            ipv4_mapping = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF'
            family: Literal[4, 6] = 4
            addr_str: str = ''
            if ipaddr[0:12] == ipv4_mapping:
                addr_str = f"{ipaddr[12]}.{ipaddr[13]}.{ipaddr[14]}.{ipaddr[15]}"
            else:
                # TODO: make ipv6 nicer
                addr_str = str(ipaddr)
                family = 6
            port = self.get_be_uint16() # TODO: i randomly assumed this would work
            servers.append(MastersrvAddr(family, addr_str, int(port)))
        return servers

    def get_uint8(self) -> int:
        res = self.byte()
        self.idx += 1
        return res

    def get_be_uint16(self) -> int:
        left = self.byte()
        self.idx += 1
        right = self.byte()
        self.idx += 1
        return left << 8 | right

    def get_optional_int(self) -> Optional[int]:
        if self.idx >= len(self._data):
            return None
        return self.get_int()

    def get_int(self) -> int:
        sign = (self.byte() >> 6) & 1
        res = self.byte() & 0x3F
        # fake loop should only loop once
        # its the poor mans goto hack
        while True:
            if (self.byte() & 0x80) == 0:
                break
            self.idx += 1
            res |= (self.byte() & 0x7F) << 6

            if (self.byte() & 0x80) == 0:
                break
            self.idx += 1
            res |= (self.byte() & 0x7F) << (6 + 7)

            if (self.byte() & 0x80) == 0:
                break
            self.idx += 1
            res |= (self.byte() & 0x7F) << (6 + 7 + 7)

            if (self.byte() & 0x80) == 0:
                break
            self.idx += 1
            res |= (self.byte() & 0x7F) << (6 + 7 + 7 + 7)
            break

        self.idx += 1
        res ^= -sign
        return res

    def get_optional_str(self, sanitize: int = SANITIZE) -> Optional[str]:
        if self.idx >= len(self._data):
            return None
        return self.get_str(sanitize)

    # TODO: optimize performance
    #       I am highly confident iterating byte by byte is very
    #       expensive in python
    #       and something common as byte filtering has to have
    #       a fast alternative
    #
    #       If there is nothing from the python standard
    #       this might be worth writing in Cython
    #       external C or rust
    # TODO: make literals work currently passing in SANITIZE as arg
    #       to unpacker.get_str() does not work because its not a literal ._.
    # def get_str(self, sanitize: Literal[0,1,2,3] = 1) -> str:
    def get_str(self, sanitize: int = SANITIZE) -> str:
        str_end: int = self.data().find(b'\x00')
        res: bytes = self.data()[:str_end]
        self.idx += str_end + 1
        if sanitize == NO_SANITIZE:
            return res.decode('utf-8', 'ignore')
        if sanitize == SANITIZE:
            return bytes( \
                [x if x > 32 or x in (9, 10, 13) else 32 for x in res]) \
                .decode('utf-8', 'ignore' \
            )
        if sanitize == SANITIZE_CC:
            return bytes( \
                [x if x > 32 else 32 for x in res]) \
                .decode('utf-8', 'ignore' \
            )
        if sanitize == SKIP_START_WHITESPACES:
            return res.decode('utf-8').lstrip()
        raise ValueError(f"Error: invalid sanitize mode {sanitize}")

# TODO: optimize performance and benchmark in tests
def pack_int(num: int) -> bytes:
    res: bytearray = bytearray(b'\x00')
    if num < 0:
        res[0] |= 0x40 # set sign bit
        num = ~num

    res[0] |= num & 0x3F # pack 6bit into res
    num >>= 6 # discard 6 bits

    i = 0
    while num != 0:
        res[i] |= 0x80 # set extend bit
        i += 1
        res.extend(bytes([num & 0x7F])) # pack 7 bit
        num >>= 7 # discard 7 bits
    return bytes(res)

def pack_be_uint16(num: int) -> bytes:
    high = (num >> 8) & 0xff
    low = num & 0xff
    return bytes([high, low])

def pack_uint8(num: int) -> bytes:
    return bytes([num])

def pack_packed_addresses(servers: list[MastersrvAddr]) -> bytes:
    res: bytes = b''
    for server in servers:
        if server.family == 4:
            ipv4_mapping = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF'
            res += ipv4_mapping
            res += bytes([int(x) for x in server.ipaddr.split('.')])
        else:
            # TODO: ipv6 is wrong
            res += server.ipaddr.encode('utf-8')
        res += pack_be_uint16(server.port)
    return res

def pack_str(data: str) -> bytes:
    return data.encode('utf-8') + b'\x00'

# TODO: optimize performance and benchmark in tests
def unpack_int(data: bytes) -> int:
    sign = (data[0] >> 6) & 1
    res = data[0] & 0x3F
    i = 0
    # fake loop should only loop once
    # its the poor mans goto hack
    while True:
        if (data[i] & 0x80) == 0:
            break
        i += 1
        res |= (data[i] & 0x7F) << 6

        if (data[i] & 0x80) == 0:
            break
        i += 1
        res |= (data[i] & 0x7F) << (6 + 7)

        if (data[i] & 0x80) == 0:
            break
        i += 1
        res |= (data[i] & 0x7F) << (6 + 7 + 7)

        if (data[i] & 0x80) == 0:
            break
        i += 1
        res |= (data[i] & 0x7F) << (6 + 7 + 7 + 7)
        break

    res ^= -sign
    return res



# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjGameDataRace(PrettyPrint):
    def __init__(
            self,
            best_time: int = 0,
            precision: int = 0,
            race_flags: int = 0
    ) -> None:
        self.item_name: str = 'item.game_data_race'
        self.type_id: int = 24
        self.id: int = 0
        self.size: int = 3

        self.best_time: int = best_time
        self.precision: int = precision
        self.race_flags: int = race_flags

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'best_time', self.best_time
        yield 'precision', self.precision
        yield 'race_flags', self.race_flags

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'best_time': self.best_time,
            'precision': self.precision,
            'race_flags': self.race_flags
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.best_time = unpacker.get_int()
        self.precision = unpacker.get_int()
        self.race_flags = unpacker.get_int() # TODO: this is a flag
        return True

    def pack(self) -> bytes:
        return pack_int(self.best_time) + \
            pack_int(self.precision) + \
            pack_int(self.race_flags)


# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int
from typing import Annotated

class ObjClientInfo(PrettyPrint):
    def __init__(
            self,
            name: Annotated[list[int], 4] = \
                [0, 0, 0, 0],
            clan: Annotated[list[int], 3] = \
                [0, 0, 0],
            country: int = 0,
            skin: Annotated[list[int], 6] = \
                [0, 0, 0, 0, 0, 0],
            use_custom_color: int = 0,
            color_body: int = 0,
            color_feet: int = 0
    ) -> None:
        self.item_name: str = 'item.client_info'
        self.type_id: int = 11
        self.id: int = 0
        self.size: int = 7

        self.name = name
        self.clan = clan
        self.country: int = country
        self.skin = skin
        self.use_custom_color: int = use_custom_color
        self.color_body: int = color_body
        self.color_feet: int = color_feet

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'name', self.name
        yield 'clan', self.clan
        yield 'country', self.country
        yield 'skin', self.skin
        yield 'use_custom_color', self.use_custom_color
        yield 'color_body', self.color_body
        yield 'color_feet', self.color_feet

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'name': self.name,
            'clan': self.clan,
            'country': self.country,
            'skin': self.skin,
            'use_custom_color': self.use_custom_color,
            'color_body': self.color_body,
            'color_feet': self.color_feet
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        for i in range(0, 4):
            self.name[i] = unpacker.get_int()
        for i in range(0, 3):
            self.clan[i] = unpacker.get_int()
        self.country = unpacker.get_int()
        for i in range(0, 6):
            self.skin[i] = unpacker.get_int()
        self.use_custom_color = unpacker.get_int()
        self.color_body = unpacker.get_int()
        self.color_feet = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return b''.join([pack_int(x) for x in self.name]) + \
            b''.join([pack_int(x) for x in self.clan]) + \
            pack_int(self.country) + \
            b''.join([pack_int(x) for x in self.skin]) + \
            pack_int(self.use_custom_color) + \
            pack_int(self.color_body) + \
            pack_int(self.color_feet)


# generated by scripts/generate_messages.py
from typing import Optional

import twnet_parser.msg6
from twnet_parser.connless_message import ConnlessMessage

import twnet_parser.messages6.connless.request_list as \
       connless_request_list
import twnet_parser.messages6.connless.list as \
       connless_list
import twnet_parser.messages6.connless.request_count as \
       connless_request_count
import twnet_parser.messages6.connless.count as \
       connless_count
import twnet_parser.messages6.connless.request_info as \
       connless_request_info
import twnet_parser.messages6.connless.info as \
       connless_info
import twnet_parser.messages6.connless.heartbeat as \
       connless_heartbeat
import twnet_parser.messages6.connless.forward_check as \
       connless_forward_check
import twnet_parser.messages6.connless.forward_response as \
       connless_forward_response
import twnet_parser.messages6.connless.forward_ok as \
       connless_forward_ok
import twnet_parser.messages6.connless.forward_error as \
       connless_forward_error

def match_connless6(msg_id: bytes, data: bytes) -> ConnlessMessage:
    msg: Optional[ConnlessMessage] = None

    if msg_id == twnet_parser.msg6.CONNLESS_REQUEST_LIST:
        msg = connless_request_list.MsgRequestList()
    elif msg_id == twnet_parser.msg6.CONNLESS_LIST:
        msg = connless_list.MsgList()
    elif msg_id == twnet_parser.msg6.CONNLESS_REQUEST_COUNT:
        msg = connless_request_count.MsgRequestCount()
    elif msg_id == twnet_parser.msg6.CONNLESS_COUNT:
        msg = connless_count.MsgCount()
    elif msg_id == twnet_parser.msg6.CONNLESS_REQUEST_INFO:
        msg = connless_request_info.MsgRequestInfo()
    elif msg_id == twnet_parser.msg6.CONNLESS_INFO:
        msg = connless_info.MsgInfo()
    elif msg_id == twnet_parser.msg6.CONNLESS_HEARTBEAT:
        msg = connless_heartbeat.MsgHeartbeat()
    elif msg_id == twnet_parser.msg6.CONNLESS_FORWARD_CHECK:
        msg = connless_forward_check.MsgForwardCheck()
    elif msg_id == twnet_parser.msg6.CONNLESS_FORWARD_RESPONSE:
        msg = connless_forward_response.MsgForwardResponse()
    elif msg_id == twnet_parser.msg6.CONNLESS_FORWARD_OK:
        msg = connless_forward_ok.MsgForwardOk()
    elif msg_id == twnet_parser.msg6.CONNLESS_FORWARD_ERROR:
        msg = connless_forward_error.MsgForwardError()

    if msg is None:
        raise ValueError(
            f"Error: unknown connless message id={msg_id!r} data={data[0]}"
        )

    msg.unpack(data)
    return msg



# generated by scripts/generate_messages.py

from typing import Any, Dict

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_int

class ObjGameDataFlag(PrettyPrint):
    def __init__(
            self,
            flag_carrier_red: int = 0,
            flag_carrier_blue: int = 0,
            flag_drop_tick_red: int = 0,
            flag_drop_tick_blue: int = 0
    ) -> None:
        self.item_name: str = 'item.game_data_flag'
        self.type_id: int = 8
        self.id: int = 0
        self.size: int = 4

        self.flag_carrier_red: int = flag_carrier_red
        self.flag_carrier_blue: int = flag_carrier_blue
        self.flag_drop_tick_red: int = flag_drop_tick_red
        self.flag_drop_tick_blue: int = flag_drop_tick_blue

    def __iter__(self):
        yield 'item_name', self.item_name
        yield 'type_id', self.type_id
        yield 'id', self.id
        yield 'size', self.size

        yield 'flag_carrier_red', self.flag_carrier_red
        yield 'flag_carrier_blue', self.flag_carrier_blue
        yield 'flag_drop_tick_red', self.flag_drop_tick_red
        yield 'flag_drop_tick_blue', self.flag_drop_tick_blue

    def to_dict_payload_only(self) -> Dict[str, Any]:
        return {
            'flag_carrier_red': self.flag_carrier_red,
            'flag_carrier_blue': self.flag_carrier_blue,
            'flag_drop_tick_red': self.flag_drop_tick_red,
            'flag_drop_tick_blue': self.flag_drop_tick_blue
        }

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker) -> bool:
        self.flag_carrier_red = unpacker.get_int()
        self.flag_carrier_blue = unpacker.get_int()
        self.flag_drop_tick_red = unpacker.get_int()
        self.flag_drop_tick_blue = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.flag_carrier_red) + \
            pack_int(self.flag_carrier_blue) + \
            pack_int(self.flag_drop_tick_red) + \
            pack_int(self.flag_drop_tick_blue)


from typing import Literal
from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker

import twnet_parser.msg6

"""
Sent by the client to ack the connect accept message CtrlAcceptConnect

Control message id 3 is only valid in 0.6.4 and earlier
got removed in 0.6.5
"""
class CtrlAckAcceptConnect(PrettyPrint):
    def __init__(self) -> None:
        self.message_type: Literal['control'] = 'control'
        self.message_name: str = 'ack_accept_connect'
        self.message_id: int = twnet_parser.msg6.CTRL_ACCEPT

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

    def unpack(self, unpacker: Unpacker, we_are_a_client: bool = True) -> bool:
        return False

    def pack(self, we_are_a_client: bool = True) -> bytes:
        return b''



from typing import Optional, Literal

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.packer import pack_str

class CtrlClose(PrettyPrint):
    def __init__(
            self,
            reason: Optional[str] = None
    ) -> None:
        self.message_type: Literal['control'] = 'control'
        self.message_name: str = 'close'
        self.message_id: int = 4

        self.reason: Optional[str] = reason

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'message_id', self.message_id

        yield 'reason', self.reason

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, unpacker: Unpacker, we_are_a_client: bool = True) -> bool:
        self.reason = unpacker.get_optional_str()
        return True

    def pack(self, we_are_a_client: bool = True) -> bytes:
        if self.reason:
            return pack_str(self.reason)
        return b''



# generated by scripts/generate_messages.py
from typing import Optional

import twnet_parser.msg6
from twnet_parser.net_message import NetMessage
from twnet_parser.ddnet_uuid import MsgDDNetUuid

import twnet_parser.messages6.system.info as \
       system6_info
import twnet_parser.messages6.system.map_change as \
       system6_map_change
import twnet_parser.messages6.system.map_data as \
       system6_map_data
import twnet_parser.messages6.system.con_ready as \
       system6_con_ready
import twnet_parser.messages6.system.snap as \
       system6_snap
import twnet_parser.messages6.system.snap_empty as \
       system6_snap_empty
import twnet_parser.messages6.system.snap_single as \
       system6_snap_single
import twnet_parser.messages6.system.input_timing as \
       system6_input_timing
import twnet_parser.messages6.system.rcon_auth_status as \
       system6_rcon_auth_status
import twnet_parser.messages6.system.rcon_line as \
       system6_rcon_line
import twnet_parser.messages6.system.ready as \
       system6_ready
import twnet_parser.messages6.system.enter_game as \
       system6_enter_game
import twnet_parser.messages6.system.input as \
       system6_input
import twnet_parser.messages6.system.rcon_cmd as \
       system6_rcon_cmd
import twnet_parser.messages6.system.rcon_auth as \
       system6_rcon_auth
import twnet_parser.messages6.system.request_map_data as \
       system6_request_map_data
import twnet_parser.messages6.system.ping as \
       system6_ping
import twnet_parser.messages6.system.ping_reply as \
       system6_ping_reply
import twnet_parser.messages6.system.rcon_cmd_add as \
       system6_rcon_cmd_add
import twnet_parser.messages6.system.rcon_cmd_remove as \
       system6_rcon_cmd_remove

def match_system6(msg_id: int, data: bytes) -> NetMessage:
    msg: Optional[NetMessage] = None

    if msg_id == twnet_parser.msg6.INFO:
        msg = system6_info.MsgInfo()
    elif msg_id == twnet_parser.msg6.MAP_CHANGE:
        msg = system6_map_change.MsgMapChange()
    elif msg_id == twnet_parser.msg6.MAP_DATA:
        msg = system6_map_data.MsgMapData()
    elif msg_id == twnet_parser.msg6.CON_READY:
        msg = system6_con_ready.MsgConReady()
    elif msg_id == twnet_parser.msg6.SNAP:
        msg = system6_snap.MsgSnap()
    elif msg_id == twnet_parser.msg6.SNAP_EMPTY:
        msg = system6_snap_empty.MsgSnapEmpty()
    elif msg_id == twnet_parser.msg6.SNAP_SINGLE:
        msg = system6_snap_single.MsgSnapSingle()
    elif msg_id == twnet_parser.msg6.INPUT_TIMING:
        msg = system6_input_timing.MsgInputTiming()
    elif msg_id == twnet_parser.msg6.RCON_AUTH_STATUS:
        msg = system6_rcon_auth_status.MsgRconAuthStatus()
    elif msg_id == twnet_parser.msg6.RCON_LINE:
        msg = system6_rcon_line.MsgRconLine()
    elif msg_id == twnet_parser.msg6.READY:
        msg = system6_ready.MsgReady()
    elif msg_id == twnet_parser.msg6.ENTER_GAME:
        msg = system6_enter_game.MsgEnterGame()
    elif msg_id == twnet_parser.msg6.INPUT:
        msg = system6_input.MsgInput()
    elif msg_id == twnet_parser.msg6.RCON_CMD:
        msg = system6_rcon_cmd.MsgRconCmd()
    elif msg_id == twnet_parser.msg6.RCON_AUTH:
        msg = system6_rcon_auth.MsgRconAuth()
    elif msg_id == twnet_parser.msg6.REQUEST_MAP_DATA:
        msg = system6_request_map_data.MsgRequestMapData()
    elif msg_id == twnet_parser.msg6.PING:
        msg = system6_ping.MsgPing()
    elif msg_id == twnet_parser.msg6.PING_REPLY:
        msg = system6_ping_reply.MsgPingReply()
    elif msg_id == twnet_parser.msg6.RCON_CMD_ADD:
        msg = system6_rcon_cmd_add.MsgRconCmdAdd()
    elif msg_id == twnet_parser.msg6.RCON_CMD_REMOVE:
        msg = system6_rcon_cmd_remove.MsgRconCmdRemove()

    # msg_id 0 is used by ddnet uuid messages
    # do not crash on those
    if msg is None:
        if msg_id != 0:
            raise ValueError(f"Error: unknown system message id={msg_id} data={data[0]}")
        msg = MsgDDNetUuid()

    msg.unpack(data)
    return msg



from typing import Literal

class MastersrvAddr():
    def __init__(
            self,
            family: Literal[4, 6] = 4,
            ipaddr: str = '127.0.0.1',
            port: int = 8303
        ) -> None:
        self.family: Literal[4, 6] = family
        self.ipaddr: str = ipaddr
        self.port: int = port




# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import SANITIZE_CC, pack_str
from typing import Literal, Optional

class MsgClCallVote(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            type: str = 'default',
            value: str = 'default',
            reason: str = 'default'
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'cl_call_vote'
        self.system_message: bool = False
        self.message_id: int = 25
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.6')
        self.header: ChunkHeader = chunk_header

        self.type: str = type
        self.value: str = value
        self.reason: str = reason

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'type', self.type
        yield 'value', self.value
        yield 'reason', self.reason

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.type = unpacker.get_str(SANITIZE_CC)
        self.value = unpacker.get_str(SANITIZE_CC)
        self.reason = unpacker.get_str(SANITIZE_CC)
        return True

    def pack(self) -> bytes:
        return pack_str(self.type) + \
            pack_str(self.value) + \
            pack_str(self.reason)